# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Main bdbcontrib API.

The bdbcontrib module serves a sandbox for experimental and semi-stable
features that are not yet ready for integreation to the bayeslite repository.
"""

from version import __version__
from py_utils import helpsub
from population import Population

@helpsub("__population_doc__", Population.__init__.__doc__)
def quickstart(*args, **kwargs):
  """__population_doc__"""
  return Population(*args, **kwargs)

__all__ = [
    'quickstart',
    'Population',
    '__version__',
]



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math

from bayeslite.shell.hook import bayesdb_shell_init

def linfoot(x):
    return math.sqrt(1. - math.exp(-2.*x))

@bayesdb_shell_init
def register_bql_math(self):
    # XXX Should not touch internals of the bdb like this -- should
    # invent a BQL API for defining BQL functions.
    self._bdb._sqlite3.createscalarfunction('exp', math.exp, 1)
    self._bdb._sqlite3.createscalarfunction('sqrt', math.sqrt, 1)
    self._bdb._sqlite3.createscalarfunction('pow', pow, 2)
    self._bdb._sqlite3.createscalarfunction('log', math.log, 1)
    self._bdb._sqlite3.createscalarfunction('linfoot', linfoot, 1)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Speed up BDB queries by parallelizing them.

Intuition
---------

``ESTIMATE SIMILARITY FROM PAIRWISE t`` will run in ``O(n^2 m v)`` time, where
``n`` is the number of rows, ``m`` is the number of models, and ``v`` is the
average number of views per model. While this query is reasonable for small
datasets, on medium-large datasets the query slows down significantly and can
become intractable.  Splitting the processing up among multiple cores can
greatly reduce computation time; this module provides functionality to assist
this multiprocessing.

Currently, a multiprocessing equivalent is provided only for
``ESTIMATE PAIRWISE SIMILARITY``. In fact, this is a query that is most likely
to require multiprocessing, as datasets frequently have many more rows than
columns.

Example
-------

Following are (very) informal timing statistics with a 200 rows by 4 column
.cvs file, run on a late 2012 MacBook Pro with a 2.5 GHz 2-core Intel Core i5::

    id,one,two,three,four
    0,2,3,4,two
    1,1,5,4,three
    2,5,1,5,one
    ...
    197,0,5,0,five
    198,5,3,0,three
    199,4,5,2,three

After inserting this .csv data into a table ``t`` and analyzing it quickly::

    bdb.execute('''
        CREATE GENERATOR t_cc FOR t USING crosscat (
            GUESS(*),
            id IGNORE
        )
    ''')
    bdb.execute('INITIALIZE 3 MODELS FOR t_cc')
    bdb.execute('ANALYZE t_cc MODELS 0-2 FOR 10 ITERATIONS WAIT')

The corresponding similarity table thus has 200^2 = 40000 rows::

    In [72]: %timeit -n 10 cursor_to_df(bdb.execute('ESTIMATE SIMILARITY FROM PAIRWISE t_cc'))
    10 loops, best of 3: 9.56 s per loop

    In [73]: %timeit -n 10 parallel.estimate_pairwise_similarity(bdb_file.name, 't', 't_cc', overwrite=True)
    10 loops, best of 3: 5.16 s per loop  # And values are located in the t_similarity table.

The approximate 2x speed up is what would be expected from dividing the work
among two cores. Further speed increases are likely with more powerful
machines.

----
"""

from bayeslite.exception import BayesLiteException as BLE
from bdbcontrib.bql_utils import cursor_to_df
import multiprocessing as mp
from bayeslite import bayesdb_open, bql_quote_name
from bayeslite.util import cursor_value


def _query_into_queue(query_string, params, queue, bdb_file):
    """
    Estimate pairwise similarity of a certain subset of the bdb according to
    query_string; place it in the multiprocessing Manager.Queue().

    For two technical reasons, this function is defined as a toplevel class and
    independently creates a bdb handle:

    1) Multiprocessing workers must be pickleable, and thus must be
       declared as toplevel functions;
    2) Multiple threads cannot access the same bdb handle, lest concurrency
       issues arise with corrupt data.

    Parameters
    ----------
    query_string : str
        Name of the query to execute, determined by estimate_similarity_mp.
    queue : multiprocessing.Manager.Queue
        Queue to place results into
    bdb_file : str
        File location of the BayesDB database. This function will
        independently open a new BayesDB handler.
    """
    bdb = bayesdb_open(pathname=bdb_file)
    res = bdb.execute(query_string, params)
    queue.put(cursor_to_df(res))


def _chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in xrange(0, len(l), n):
        yield l[i:i+n]


def estimate_pairwise_similarity(bdb_file, table, model, sim_table=None,
                                 cores=None, N=None, overwrite=False):
    """
    Estimate pairwise similarity from the given model, splitting processing
    across multiple processors, and save results into sim_table.

    Because called methods in this function must also open up separate BayesDB
    instances, this function accepts a BayesDB filename, rather than an actual
    bayeslite.BayesDB object.

    Parameters
    ----------
    bdb_file : str
        File location of the BayesDB database object. This function will
        handle opening the file with bayeslite.bayesdb_open.
    table : str
        Name of the table containing the raw data.
    model : str
        Name of the metamodel to estimate from.
    sim_table : str
        Name of the table to insert similarity results into. Defaults to
        table name + '_similarity'.
    cores : int
        Number of processors to use. Defaults to the number of cores as
        identified by multiprocessing.num_cores.
    N : int
        Number of rows for which to estimate pairwise similarities (so
        N^2 calculations are done). Should be used just to test small
        batches; currently, there is no control over what specific pairwise
        similarities are estimated with this parameter.
    overwrite : bool
        Whether to overwrite the sim_table if it already exists. If
        overwrite=False and the table exists, function will raise
        sqlite3.OperationalError. Default True.
    """
    bdb = bayesdb_open(pathname=bdb_file)

    if cores is None:
        cores = mp.cpu_count()

    if cores < 1:
        raise BLE(ValueError(
            "Invalid number of cores {}".format(cores)))

    if sim_table is None:
        sim_table = table + '_similarity'

    # Get number of occurrences in the database
    count_cursor = bdb.execute(
        'SELECT COUNT(*) FROM {}'.format(bql_quote_name(table))
    )
    table_count = cursor_value(count_cursor)
    if N is None:
        N = table_count
    elif N > table_count:
        raise BLE(ValueError(
            "Asked for N={} rows but {} rows in table".format(N, table_count)))

    # Calculate the size (# of similarities to compute) and
    # offset (where to start) calculation for each worker query.

    # Divide sizes evenly, and make the last job finish the remainder
    sizes = [(N * N) / cores for i in range(cores)]
    sizes[-1] += (N * N) % cores

    total = 0
    offsets = [total]
    for size in sizes[:-1]:
        total += size
        offsets.append(total)

    # Create the similarity table. Assumes original table has rowid column.
    # XXX: tables don't necessarily have an autoincrementing primary key
    # other than rowid, which is implicit and can't be set as a foreign key.
    # We ought to ask for an optional user-specified foreign key, but
    # ESTIMATE SIMILARITY returns numerical values rather than row names, so
    # changing numerical rownames into that foreign key would be finicky. For
    # now, we eliminate REFERENCE {table}(foreign_key) from the rowid0 and
    # rowid1 specs.
    sim_table_q = bql_quote_name(sim_table)
    if overwrite:
        bdb.sql_execute('DROP TABLE IF EXISTS {}'.format(sim_table_q))

    bdb.sql_execute('''
        CREATE TABLE {} (
            rowid0 INTEGER NOT NULL,
            rowid1 INTEGER NOT NULL,
            value DOUBLE NOT NULL
        )
    '''.format(sim_table_q))

    # Define the helper which inserts data into table in batches
    def insert_into_sim(df):
        """
        Use the main thread bdb handle to successively insert results of
        ESTIMATEs into the table.
        """
        rows = map(list, df.values)
        insert_sql = '''
            INSERT INTO {} (rowid0, rowid1, value) VALUES (?, ?, ?)
        '''.format(sim_table_q)
        # Avoid sqlite3 500-insert limit by grouping insert statements
        # into one transaction.
        with bdb.transaction():
            for row in rows:
                bdb.sql_execute(insert_sql, row)

    pool = mp.Pool(processes=cores)

    manager = mp.Manager()
    queue = manager.Queue()

    # Construct the estimate query template.
    q_template = '''
        ESTIMATE SIMILARITY FROM PAIRWISE {} LIMIT ? OFFSET ?
    ''' .format(bql_quote_name(model))

    for so in zip(sizes, offsets):
        pool.apply_async(
            _query_into_queue, args=(q_template, so, queue, bdb_file)
        )

    # Close pool and wait for processes to finish
    # FIXME: This waits for all processes to finish before inserting
    # into the table, which means that memory usage is potentially very
    # high!
    pool.close()
    pool.join()

    # Process returned results
    while not queue.empty():
        df = queue.get()
        insert_into_sim(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Analyze and save .bdb files for Satellites.

This script simultaneously serves two purposes:
- To prepare an analyzed .bdb file for distribution to clients of the
  Satellites demo
- To prepare a series of .bdb files for consumption by the stability
  probing script probe.py.

Each produced file is named with a 'satellites' prefix.  The file name
additionally captures
- a date stamp,
- the running user,
- the number of models [analysis snapshots only], and
- the number of analysis iterations [analysis snapshots only].

For each foo.bdb, this program saves a foo-meta.txt containing the
following information:
- name and sha256sum of the file described;
- # models; # iterations; time taken; initial entropy; parallelism
  level; date stamp; user stamp;
- crosscat version; bayeslite version; bdbcontrib version
  (including a full copy of this driver script); and
- logscore history plot [distributable .bdb only].

"""

# Speed rules of thumb:
# - 64 models and 1500 iterations took ~30 minutes on probcomp in late September
# - 64 * 5 models and 300 iterations took ~18 minutes on probcomp 10/5/15
# - 64 * 50 models and 30 iterations took ~36 minutes on probcomp 10/5/15

import argparse
import datetime
import hashlib
import logging
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os
import subprocess
import sys
import time

import bayeslite
import bayeslite.metamodels.crosscat
import bdbcontrib
import bdbcontrib.crosscat_utils
import bdbcontrib.bql_utils
import crosscat
import crosscat.MultiprocessingEngine as ccme
from bdbcontrib.experiments.probe import log

def doit(out_dir, num_models, num_iters, checkpoint_freq, seed):
    then = time.time()

    timestamp = datetime.datetime.fromtimestamp(then).strftime('%Y-%m-%d')
    user = subprocess.check_output(["whoami"]).strip()
    host = subprocess.check_output(["hostname"]).strip()
    filestamp = '-' + timestamp + '-' + user
    def out_file_name(base, ext):
        return out_dir + '/' + base + filestamp + ext

    csv_file = os.path.join(os.path.dirname(__file__), 'satellites.csv')
    bdb_file = out_file_name('satellites', '.bdb')

    # so we can build bdb models
    os.environ['BAYESDB_WIZARD_MODE']='1'

    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    if os.path.exists(bdb_file):
        print 'Error: File', bdb_file, 'already exists. Please remove it.'
        sys.exit(1)

    # create database mapped to filesystem
    log('opening bdb on disk: %s' % bdb_file)
    bdb = bayeslite.bayesdb_open(pathname=bdb_file, builtin_metamodels=False)

    def execute(bql):
        log("executing %s" % bql)
        bdb.execute(bql)

    # read csv into table
    log('reading data from %s' % csv_file)
    bayeslite.bayesdb_read_csv_file(bdb, 'satellites', csv_file,
            header=True, create=True, ifnotexists=True)

    # Add a "not applicable" orbit sub-type
    log('adding "not applicable" orbit sub-type')
    bdb.sql_execute('''UPDATE satellites
        SET type_of_orbit = 'N/A'
        WHERE (class_of_orbit = 'GEO' OR class_of_orbit = 'MEO')
          AND type_of_orbit = 'NaN'
    ''')

    # nullify "NaN"
    log('nullifying NaN')
    bdbcontrib.bql_utils.nullify(bdb, 'satellites', 'NaN')

    # register crosscat metamodel
    cc = ccme.MultiprocessingEngine(seed=seed)
    ccmm = bayeslite.metamodels.crosscat.CrosscatMetamodel(cc)
    bayeslite.bayesdb_register_metamodel(bdb, ccmm)

    # create the crosscat generator using
    execute('''
        CREATE GENERATOR satellites_cc FOR satellites USING crosscat (
            GUESS(*),
            name IGNORE,
            Country_of_Operator CATEGORICAL,
            Operator_Owner CATEGORICAL,
            Users CATEGORICAL,
            Purpose CATEGORICAL,
            Class_of_Orbit CATEGORICAL,
            Type_of_Orbit CATEGORICAL,
            Perigee_km NUMERICAL,
            Apogee_km NUMERICAL,
            Eccentricity NUMERICAL,
            Period_minutes NUMERICAL,
            Launch_Mass_kg NUMERICAL,
            Dry_Mass_kg NUMERICAL,
            Power_watts NUMERICAL,
            Date_of_Launch NUMERICAL,
            Anticipated_Lifetime NUMERICAL,
            Contractor CATEGORICAL,
            Country_of_Contractor CATEGORICAL,
            Launch_Site CATEGORICAL,
            Launch_Vehicle CATEGORICAL,
            Source_Used_for_Orbital_Data CATEGORICAL,
            longitude_radians_of_geo NUMERICAL,
            Inclination_radians NUMERICAL
        )
    ''')

    execute('INITIALIZE %d MODELS FOR satellites_cc' % (num_models,))

    cur_iter_ct = 0

    def snapshot():
        log('vacuuming')
        bdb.sql_execute('vacuum')
        cur_infix = '-%dm-%di' % (num_models, cur_iter_ct)
        save_file_name = out_file_name('satellites', cur_infix + '.bdb')
        meta_file_name = out_file_name('satellites', cur_infix + '-meta.txt')
        log('recording snapshot ' + save_file_name)
        os.system("cp %s %s" % (bdb_file, save_file_name))
        report(save_file_name, meta_file_name)

    def record_metadata(f, saved_file_name, sha_sum, total_time,
                        plot_file_name=None):
        f.write("DB file " + saved_file_name + "\n")
        f.write(sha_sum)
        f.write("built from " + csv_file + "\n")
        f.write("by %s@%s\n" % (user, host))
        f.write("at seed %s\n" % seed)
        f.write("in %3.2f seconds\n" % total_time)
        f.write("with %s models analyzed for %s iterations\n"
                % (num_models, num_iters))
        f.write("by bayeslite %s, with crosscat %s and bdbcontrib %s\n"
                % (bayeslite.__version__, crosscat.__version__, bdbcontrib.__version__))
        if plot_file_name is not None:
            f.write("diagnostics recorded to %s\n" % plot_file_name)
        f.flush()

    def report(saved_file_name, metadata_file, echo=False, plot_file_name=None):
        sha256 = hashlib.sha256()
        with open(saved_file_name, 'rb') as fd:
            for chunk in iter(lambda: fd.read(65536), ''):
                sha256.update(chunk)
        sha_sum = sha256.hexdigest() + '\n'
        total_time = time.time() - then
        with open(metadata_file, 'w') as fd:
            record_metadata(fd, saved_file_name,
                            sha_sum, total_time, plot_file_name)
            fd.write('using script ')
            fd.write('-' * 57)
            fd.write('\n')
            fd.flush()
            os.system("cat %s >> %s" % (__file__, metadata_file))

        if echo:
            record_metadata(sys.stdout, saved_file_name,
                            sha_sum, total_time, plot_file_name)

    def final_report():
        # create a diagnostics plot
        plot_file_name = out_file_name('satellites', '-logscores.pdf')
        log('writing diagnostic plot to %s' % plot_file_name)
        _fig = bdbcontrib.crosscat_utils.plot_crosscat_chain_diagnostics(
            bdb, 'logscore', 'satellites_cc')
        plt.savefig(plot_file_name)
        final_metadata_file = out_file_name('satellites', '-meta.txt')
        report(bdb_file, final_metadata_file,
               echo=True, plot_file_name=plot_file_name)

    snapshot()
    while cur_iter_ct < num_iters:
        execute('ANALYZE satellites_cc FOR %d ITERATIONS WAIT' % checkpoint_freq)
        cur_iter_ct += checkpoint_freq
        snapshot()

    final_report()

    log('closing bdb %s' % bdb_file)
    bdb.close()
    os.system("cd %s && ln -s satellites%s.bdb satellites.bdb" % (out_dir, filestamp))

parser = argparse.ArgumentParser(
    description="Analyze and save .bdb files for Satellites.")
parser.add_argument(
    '-o', '--outdir', default="output",
    help="Directory to save generated .bdb files [default: \"output\"]")
parser.add_argument(
    '-m', '--models', type=int,
    help="Number of models to analyze")
parser.add_argument(
    '-i', '--iters', type=int,
    help="Number of iterations of analysis to run")
parser.add_argument(
    '-c', '--checkpoint_freq', type=int,
    help="Frequency of checkpoints to take [default: no checkpoints]")
parser.add_argument(
    '-s', '--seed', type=int, default=0,
    help="Initial entropy [default: 0]")

def main():
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)
    doit(args.outdir, args.models, args.iters, args.checkpoint_freq, args.seed)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import copy
import glob
import os
import shutil
import sys
import tempfile

root = os.path.dirname(os.path.dirname(__file__))
sats_path = os.path.join(root, "examples", "satellites")

def test_stability_check_integration_smoke__ci_because_43s_is_too_slow():
    bdbs_dir = None
    plots_dir = None
    old_path = copy.copy(sys.path)
    sys.path.append(sats_path)
    try:
        import build_bdbs
        import probe
        import visualize
        bdbs_dir = tempfile.mkdtemp(suffix='bdbs')
        plots_dir = tempfile.mkdtemp(suffix='plots')
        build_bdbs.doit(
            bdbs_dir, num_models=4, num_iters=2, checkpoint_freq=1, seed=0)
        with tempfile.NamedTemporaryFile() as f:
            probe.doit(glob.glob("%s/*i.bdb" % bdbs_dir),
                       f.name, model_schedule=[1,2], n_replications=2)
            visualize.plot_probe_results(f.name, plots_dir)
    finally:
        sys.path = old_path
        if bdbs_dir is not None:
            shutil.rmtree(bdbs_dir)
        if plots_dir is not None:
            shutil.rmtree(plots_dir)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json

import numpy as np
import matplotlib
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
import seaborn as sns

import bayeslite.core
from bayeslite.exception import BayesLiteException as BLE
from crosscat.utils import sample_utils as su

from population_method import population_method
import bql_utils as bu
import plot_utils as pu


###############################################################################
###                                 PUBLIC                                  ###
###############################################################################

@population_method(population_to_bdb=0, generator_name=1)
def draw_crosscat(bdb, generator, modelno, row_label_col=None):
    """Draw crosscat model from the specified generator.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    generator_name : str
        Name of generator.
    modelno: int
        Number of model to draw.

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    bql = '''
        SELECT tabname, metamodel FROM bayesdb_generator
            WHERE name = ?
    '''
    cursor = bdb.execute(bql, (generator,))
    table_name, metamodel = cursor.next()

    if metamodel.lower() != 'crosscat':
        raise BLE(ValueError(
            'Metamodel for generator %s (%s) should be crosscat' %
            (generator, metamodel)))

    figure, axes = plt.subplots(tight_layout=False)
    draw_state(bdb, table_name, generator,
        modelno, ax=axes, row_label_col=row_label_col)

    return figure

@population_method(population_to_bdb=0, generator_name=2)
def plot_crosscat_chain_diagnostics(bdb, diagnostic, generator):
    """Plot diagnostics for all models of generator.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    diagnostic : str
        Valid (crosscat) diagnostics are:
            - logscore: log score of the model
            - num_views: the number of views in the model
            - column_crp_alpha: CRP alpha over columns
    generator : str
        Name of the generator to diagnose.

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    valid_diagnostics = ['logscore', 'num_views', 'column_crp_alpha']
    if diagnostic not in valid_diagnostics:
        raise BLE(ValueError('Unknown diagnostic %s.\n'
            'Please choose one of the following instead: %s\n'
            % ', '.join(valid_diagnostics)))

    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator)

    # Get model numbers. Do not rely on there to be a diagnostic for every
    # model.
    bql = '''
        SELECT modelno, COUNT(modelno) FROM bayesdb_crosscat_diagnostics
            WHERE generator_id = ? GROUP BY modelno
    '''
    df = bu.cursor_to_df(bdb.execute(bql, (generator_id,)))
    models = df['modelno'].astype(int).values

    figure, ax = plt.subplots(tight_layout=True, figsize=(10, 5))
    colors = sns.color_palette("GnBu_d", len(models))
    for i, modelno in enumerate(models):
        bql = '''
            SELECT {}, iterations FROM bayesdb_crosscat_diagnostics
                WHERE modelno = ? AND generator_id = ?
                ORDER BY iterations ASC
        '''.format(diagnostic)
        df = bu.cursor_to_df(bdb.execute(bql, (modelno, generator_id)))
        ax.plot(df['iterations'].values, df[diagnostic].values,
                 c=colors[modelno], alpha=.7, lw=2)

        ax.text(df['iterations'].values[-1], df[diagnostic].values[-1],
                str(modelno), color=colors[i])

    ax.set_xlabel('Iteration')
    ax.set_ylabel(diagnostic)
    ax.set_title('%s for each model in %s' % (diagnostic, generator,))

    return figure

###############################################################################
###                              INTERNAL                                   ###
###############################################################################

def get_cols_in_view(X_L, view):
    return [c for c, v in enumerate(X_L['column_partition']['assignments'])
        if v == view]


def get_rows_in_cluster(X_D, view, cluster):
    return [r for r, c in enumerate(X_D[view]) if c == cluster]


def get_M_c(bdb, generator_name):
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator_name)
    sql = '''
        SELECT metadata_json FROM bayesdb_crosscat_metadata
            WHERE generator_id = ?
    '''
    cursor = bdb.sql_execute(sql, (generator_id,))
    try:
        row = cursor.next()
    except StopIteration:
        raise BLE(ValueError(bdb, 'No crosscat metadata for generator: %s'
            % (generator_name,)))
    else:
        return json.loads(row[0])


def get_metadata(bdb, generator_name, modelno):
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator_name)
    sql = '''
        SELECT theta_json FROM bayesdb_crosscat_theta
            WHERE generator_id = ? and modelno = ?
    '''
    cursor = bdb.sql_execute(sql, (generator_id, modelno))
    try:
        row = cursor.next()
    except StopIteration:
        raise BLE(ValueError('Could not find generator with '
            'name {}, or incorrect model number.'.format(generator_name)))
    else:
        return json.loads(row[0])


def get_row_probabilities(X_L, X_D, M_c, T, view):
    """Returns predictive probability of the data in each row of T in view."""
    num_rows = len(X_D[0])
    cols_in_view = get_cols_in_view(X_L, view)
    num_clusters = max(X_D[view])+1
    cluster_models = [su.create_cluster_model_from_X_L(M_c, X_L, view, c)
        for c in range(num_clusters)]

    logps = np.zeros(num_rows)
    for row in range(num_rows):
        cluster_idx = X_D[view][row]
        for col in cols_in_view:
            x = T[row][col]
            if x is None:
                continue
            elif isinstance(x, float):
                if np.isnan(x):
                    continue
            else:
                x = M_c['column_metadata'][col]['code_to_value'][x]
            component_model = cluster_models[cluster_idx][col]
            component_model.remove_element(x)
            logps[row] += component_model.calc_element_predictive_logp(x)
            component_model.insert_element(x)

    assert len(logps) == len(X_D[0])
    return logps


def get_column_probabilities(X_L, M_c):
    """Returns marginal probability of each column."""
    num_cols = len(X_L['column_partition']['assignments'])
    num_views = len(X_L['view_state'])
    logps = np.zeros(num_cols)
    for view in range(num_views):
        num_clusters = len(
            X_L['view_state'][view]['row_partition_model']['counts'])
        for cluster in range(num_clusters):
            cluster_model = su.create_cluster_model_from_X_L(M_c, X_L, view,
                cluster)
            for col in get_cols_in_view(X_L, view):
                component_model = cluster_model[col]
                logps[col] += component_model.calc_marginal_logp()
    return logps


def draw_state(bdb, table_name, generator_name, modelno,
        ax=None, border_width=3, row_label_col=None, short_names=True,
        hilight_rows=[], hilight_rows_colors=None,
        hilight_cols=[], hilight_cols_colors=None,
        separator_color='black', separator_width=4,
        blank_state=False, nan_color=(1., 0., 0., 1.),
        view_labels=None, view_label_fontsize='large',
        legend=True, legend_fontsize='medium',
        row_legend_loc=1, row_legend_title='Row key',
        col_legend_loc=4, col_legend_title='Column key',
        descriptions_in_legend=True, legend_wrap_threshold=20,):
    """Creates a debugging (read: not pretty) rendering of a CrossCat state.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        The BayesDB object associated with the CrossCat state
    table_name : str
        The btable name containing the data associated with the state
    generator_name : str
        The CrossCat generator associated witht the state
    modelno : int
        The index of the model/state to draw
    ax : matplotlib.axis
        The axis on which to draw
    row_label_col : str
        The name of the column to use for row labels. Defaults to FIXME
    short_names : bool
        Use shortnames as column labels
    hilight_rows : list<str>
        A list of rows to hilight with colored rectangles.
    hilight_rows_colors : list
        Contains a color (str or tuple) for each entry in `hilight_rows`. If
        not specified, unique colors for each entry are generated.
    hilight_cols : list
        A list of columns to hilight with colored rectangles.
    hilight_cols_colors : list
        Contains a color (str or tuple) for each entry in `hilight_cols`. If
        not specified, unique colors for each entry are generated.
    blank_state : bool
        If True, draws an unsorted, unpartitioned state
    view_labels : list<str>
        Labels placed above each view. If `len(view_labels) < num_views` then
        only the views for which there are entries are labeled.
    view_label_fontsize : valid matplotlib `fontsize`
        Font size used for vie labels
    legend : bool
        If True (defult) displays legend
    legend_fontsize : valid matplotlib `fontsize`
        Font size used for legend entries and titles
    row_legend_loc : matplotlib.legend location
        location of the row legend. For use with row hilighting
    col_legend_loc : matplotlib.legend location
        location of the column legend. For use with column hilighting

    Returns
    -------
    ax : matplotlib.axis
        The state rendering

    Other Parameters
    ----------------
    border_width : int
        The number of cells between views. Use larger values for longer row
        names.
    separator_color : str or (r, g, b) or (r, g, b, alpha) tuple
        The color of the cluster seprator. Default is black.
    separator_width : int
        linewidth of the cluster separator
    nan_color : str or (r, g, b) or (r, g, b, alpha) tuple
        The color for missing/NaN values. Default is red.
    row_legend_title : str
        title of the row legend
    col_legend_title : str
        title of the column legend
    legend_wrap_threshold : int
        Max number of characters until wordrap for collapsed legends. For use
        when multiple entries in `hilight_cols_colors` or `hilight_cols_colors`
        contain the same color.
    descriptions_in_legend : bool
        If True (default), the column descriptions (requires codebook) are
        added to the legend
    """
    theta = get_metadata(bdb, generator_name, modelno)
    M_c = get_M_c(bdb, generator_name)
    # idx_to_name doesn't use an int idx, but a string idx because
    # crosscat.  Yep.
    ordered_columns = [M_c['idx_to_name'][str(idx)] for
                       idx in sorted(M_c['name_to_idx'].values())]
    T = bu.get_data_as_list(bdb, table_name, column_list=ordered_columns)
    X_L = theta['X_L']
    X_D = theta['X_D']

    num_rows = len(T)
    num_cols = len(T[0])

    if not blank_state:
        sortedstate = DrawStateUtils.sort_state(X_L, X_D, M_c, T)
        sorted_views, sorted_clusters, sorted_cols, sorted_rows = sortedstate
        column_partition = X_L['column_partition']['assignments']
    else:
        blankstate = DrawStateUtils.gen_blank_sort(num_rows, num_cols)
        sorted_views, sorted_clusters, sorted_cols, sorted_rows = blankstate
        column_partition = [0]*num_cols

    if view_labels is not None:
        if not isinstance(view_labels, list):
            raise BLE(TypeError("view_labels must be a list"))
        if len(view_labels) != len(sorted_views):
            view_labels += ['']*(len(sorted_rows)-len(view_labels))
    else:
        view_labels = ['V ' + str(i) for i in range(num_rows)]

    if hilight_cols_colors is None:
        hilight_cols_colors = []

    if hilight_rows_colors is None:
        hilight_rows_colors = []

    # set colormap to 50% gray (should probably give the user control
    # over this)
    cmap = matplotlib.colors.ListedColormap([(1, 1, 1, 1), (1, 1, 1, 1)])
    T = DrawStateUtils.convert_t_do_numerical(T, M_c)

    num_views = len(sorted_cols)
    X = np.zeros((num_rows, num_cols+num_views*border_width))

    # row hilighting
    row_hl_colors = DrawStateUtils.gen_hilight_colors(hilight_rows,
        hilight_rows_colors)

    hl_row_idx_label_zip = []
    if row_label_col is None:
        row_labels = [str(i) for i in range(num_rows)]
    elif isinstance(row_label_col, list):
        if len(row_label_col) != num_rows:
            raise BLE(TypeError("If row_label_col is a list, it must have an "
                                "entry for each row"))
        row_labels = [str(label) for label in row_label_col]
    elif isinstance(row_label_col, str):
        # FIXME: This is not going to work until BayesDB stops removing key and
        # ignore columns from the data
        raise NotImplementedError
        label_col_idx = M_c['name_to_idx'][row_label_col]
        row_labels = [str(T[row, label_col_idx]) for row in range(num_rows)]
    else:
        raise BLE(TypeError("Unhandled row_label_col type {}.".format(
            type(row_label_col))))

    row_idx_to_label = {}
    row_label_to_idx = {}
    for row, label in enumerate(row_labels):
        # XXX: Allows missing enries to be a column label
        row_idx_to_label[row] = label
        row_label_to_idx[label] = row

    for label in hilight_rows:
        hl_row_idx_label_zip.append((row_label_to_idx[label], label,))

    # column hilighting
    col_hl_colors = DrawStateUtils.gen_hilight_colors(hilight_cols,
        hilight_cols_colors)

    hl_col_idx_label_zip = []
    for label in hilight_cols:
        hl_col_idx_label_zip.append((M_c['name_to_idx'][label], label,))

    # generate a heatmap using the data (allows clusters to ahve different
    # base colors)
    cell_colors = DrawStateUtils.gen_cell_colors(T, sorted_views, sorted_cols,
        sorted_clusters, sorted_rows, column_partition, cmap, border_width,
        nan_color=nan_color)

    # x_tick_labels = []
    x_labels = []

    if ax is None:
        ax = plt.gca()

    ax.imshow(cell_colors, cmap=cmap, interpolation='nearest', origin='upper',
              aspect='auto')
    col_count = 0
    for v, view in enumerate(sorted_views):
        view_x_labels = [M_c['idx_to_name'][str(col)]
                         for col in sorted_cols[view]]
        if short_names:
            view_x_tick_labels = bu.get_shortnames(bdb, table_name,
                view_x_labels)
        else:
            view_x_tick_labels = view_x_labels

        y_tick_labels = []

        x_labels += view_x_labels + ['_']*border_width
        num_cols_view = len(sorted_cols[view])
        sbplt_start = col_count+v*border_width
        sbplt_end = col_count+num_cols_view+v*border_width

        for i, vxtl in enumerate(view_x_labels):
            if vxtl in hilight_cols:
                edgecolor = col_hl_colors[vxtl]
                x_a = sbplt_start+i-.5
                ax.add_patch(Rectangle((x_a, -.5), 1, num_rows,
                                       facecolor="none", edgecolor=edgecolor,
                                       lw=2, zorder=10))
                fontcolor = edgecolor
                fontsize = 'x-small'
            else:
                fontcolor = '#333333'
                fontsize = 'x-small'
            font_kws = dict(color=fontcolor, fontsize=fontsize, rotation=90,
                            va='top', ha='center')
            ax.text(sbplt_start+i+.5, num_rows+.5, view_x_tick_labels[i],
                    font_kws)

        view_label_x = (sbplt_start+sbplt_end)/2. - .5
        view_label_y = -2.5
        font_kws = dict(ha='center',
                        fontsize=view_label_fontsize,
                        weight='bold')
        ax.text(view_label_x, view_label_y, view_labels[v], font_kws)

        y = 0
        for cluster in sorted_clusters[view]:
            y_tick_labels += [row_idx_to_label[row]
                              for row in sorted_rows[view][cluster]]
            ax.plot([sbplt_start-.5, sbplt_end-.5], [y-.5, y-.5],
                    color=separator_color, lw=separator_width)
            for row, label in hl_row_idx_label_zip:
                try:
                    pos = sorted_rows[view][cluster].index(row)
                except ValueError:
                    pos = None

                if pos is not None:
                    edgecolor = row_hl_colors[label]
                    ax.add_patch(Rectangle((sbplt_start - .5, y + pos - .5),
                                           num_cols_view, 1, facecolor="none",
                                           edgecolor=edgecolor, lw=2,
                                           zorder=10))

            y += len(sorted_rows[view][cluster])

        for i, row in enumerate(range(num_rows-1, -1, -1)):
            if y_tick_labels[i] in hilight_rows:
                fontcolor = row_hl_colors[y_tick_labels[i]]
                fontsize = 'x-small'
                fontweight = 'bold'
                zorder = 10
            else:
                fontsize = 'x-small'
                fontcolor = '#333333'
                fontweight = 'light'
                zorder = 5

            ax.text(sbplt_start-1, i+.5, str(y_tick_labels[i]), ha='right',
                    fontsize=fontsize, color=fontcolor, weight=fontweight,
                    zorder=zorder)
        col_count += num_cols_view

    # generate row legend
    # Use matplotlib artists to generate a list of colored lines
    # TODO: Refactor legend generator into its own function
    if legend:
        if len(hilight_rows) > 0:
            row_legend = pu.gen_collapsed_legend_from_dict(
                row_hl_colors, loc=row_legend_loc, title=row_legend_title,
                fontsize=legend_fontsize, wrap_threshold=legend_wrap_threshold)
            ax.add_artist(row_legend)

        if len(hilight_cols) > 0:
            col_legend_labels = bu.get_shortnames(bdb, table_name,
                hilight_cols)
            if descriptions_in_legend:
                for i, col_id in enumerate(hilight_cols):
                    col_legend_labels[i] += ': ' + bu.get_descriptions(
                        bdb, table_name, [col_id])[0]
                    col_legend_labels[i] = col_legend_labels[i]

            col_legend = pu.gen_collapsed_legend_from_dict(
                dict(zip(col_legend_labels, hilight_cols_colors)),
                loc=col_legend_loc, title=col_legend_title,
                fontsize=legend_fontsize, wrap_threshold=legend_wrap_threshold)

            ax.add_artist(col_legend)

    ax.tick_params(**{
        'axis': 'both',
        'length': 0
    })
    ax.set_xlim([-.5, X.shape[1]])
    ax.set_ylim([X.shape[0], -.5])
    ax.spines['bottom'].set_color('white')
    ax.spines['top'].set_color('white')
    ax.spines['right'].set_color('white')
    ax.spines['left'].set_color('white')
    ax.set_yticks(range(num_rows))
    ax.set_xticks(range(num_cols+num_views*border_width))
    ax.tick_params(axis='x', colors='white')
    # ax.set_xticklabels(x_tick_labels, rotation=90, color='black', fontsize=9)
    ax.set_yticklabels(['']*num_rows)
    ax.tick_params(axis='y', colors='white')
    ax.grid(b=False)
    ax.set_axis_bgcolor('white')
    return ax


class DrawStateUtils(object):
    # This class contains several internal helper functions for the heavyweight
    # draw_state function. All the functions are static.

    @staticmethod
    def convert_t_do_numerical(T, M_c):
        for colno, md in enumerate(M_c['column_metadata']):
            lookup = md['code_to_value']
            if len(lookup) == 0:
                continue
            for row in range(len(T)):
                val = T[row][colno]
                if val is None:
                    continue
                if isinstance(val, float):
                    if np.isnan(val):
                        continue
                T[row][colno] = lookup[val]

        Tary = np.array(T)
        return Tary


    @staticmethod
    def sort_state(X_L, X_D, M_c, T):
        """Sorts the metadata for visualization.

        Sort views from largest to smallest. W/in views, sorts columns
        r->l from highest marginal probability to lowest; and sorts
        categories largest to smallest from top to bottom. Within
        categories, sorts rows higest to lowest probability from top
        to bottom.

        Parameters
        ----------
        X_L : dict
            CrossCat view metadata
        X_D : list
            CrossCat row/view/cluster partition
        M_c : dict
            CrossCat column metadata
        T : list

        Notes
        -----
        Calculating probabilites requires initializing stateless CrossCat
        component models, which is a little exensive.

        Returns
        -------
        sorted_views : list
            view indices from largest (most columns) to smallest.
        sorted_clusters : dict<list>
            sorted_clusters[v] is a sorted list of clusters in view v from
            largest (most rows) to smallest.
        sorted_cols : dict<list>
            sorted_columns[v] is a sorted list of columns in view v from
            highest to lowest probability.
        sorted_rows : dict<dict<list>>
            sorted_rows[v][c] is a sorted list of rows in cluster c of view v
            from higest to lowest probability.
        """
        num_views = len(X_L['view_state'])

        column_logps = get_column_probabilities(X_L, M_c)

        cols_by_view = [get_cols_in_view(X_L, v) for v in range(num_views)]
        num_cols_in_view = [len(cv) for cv in cols_by_view]
        sorted_views = [i[0] for i in sorted(enumerate(num_cols_in_view),
                        reverse=True, key=lambda x:x[1])]

        sorted_cols = {}
        sorted_rows = {}
        sorted_clusters = {}
        for view in sorted_views:
            num_clusters = len(X_L['view_state'][view]['row_partition_model']['counts'])

            # get all columns in the view and sort by probability.
            col_prob = {}
            for col in cols_by_view[view]:
                logp = column_logps[col]
                # index by logp because that is how we're going to sort
                col_prob[logp] = col

            sorted_cols[view] = [col_prob[p] for p in sorted(col_prob.keys(),
                                 reverse=True)]

            # sort clusters by size
            rows_by_cluster = [get_rows_in_cluster(X_D, view, clstr) for
                               clstr in range(num_clusters)]
            num_rows_in_clstr = [len(rc) for rc in rows_by_cluster]
            sorted_clusters[view] = [
                i[0] for i in sorted(enumerate(num_rows_in_clstr), reverse=True,
                                     key=lambda x:x[1])]

            row_logps = get_row_probabilities(X_L, X_D, M_c, T, view)
            sorted_rows_view = {}
            num_rows_accounted_for = 0
            for clstr in sorted_clusters[view]:
                rows_in_clstr = get_rows_in_cluster(X_D, view, clstr)
                row_prob = [row_logps[row] for row in rows_in_clstr]
                sorted_row_clstr = [rows_in_clstr[i] for
                                    i in np.argsort(row_prob)][::-1]

                assert len(rows_in_clstr) == len(sorted_row_clstr)

                num_rows_accounted_for += len(sorted_row_clstr)
                sorted_rows_view[clstr] = sorted_row_clstr

            assert num_rows_accounted_for == len(X_D[0])

            sorted_rows[view] = sorted_rows_view

        return sorted_views, sorted_clusters, sorted_cols, sorted_rows


    @staticmethod
    def cmap_color_brightness(value, base_color, vmin, vmax,
            nan_color=(1., 0., 0., 1.)):
        """Returns a color representing the magnitude of a value.

        Higher values are represented with lighter colors. Given the a
        base color of 50% gray, white represents the max and black
        represents the min.

        Parameters
        ----------
        value : float
            The value of a cell to convert to a color
        base_color: tuple (r, g, b) or (r, g, b, alpha)
            The color to which the brightness manipulation is applied
        vmin : float
            The minimum value of the column to which `value` belongs
        vmax : float
            The maximum value of the column to which `value` belongs
        nan_color : tuple (r, g, b) or (r, g, b, alpha)
            The color to use for NaN or missing data. Defaults to red.

        Returns
        -------
            color : tuple (r, g, b, a)
                The `value` color
        """
        # XXX: bayesdb_data never retruns NaN for multinomial---NaN is added
        # to value_map
        if value is None or np.isnan(value):
            return nan_color

        if vmin == vmax:
            brightness = .5
        else:
            span = vmax - vmin
            # brightness = .5*(value-vmin)/(span)+.25  # low contrast
            brightness = (value-vmin)/(span)

        color = np.array([min(c*brightness, 1.) for c in list(base_color)])
        color[3] = 1.

        return color


    @staticmethod
    def gen_hilight_colors(hl_labels=None, hl_colors=None):
        """Generates a hilight color lookup from labels to colors. Generates
        labels from Set1 by default.
        """
        if hl_labels is None:
            return {}

        hl_colors_out = {}
        if hl_colors is None:
            if len(hl_labels) > 0:
                hl_max = float(len(hl_labels))
                hl_colors_out = dict([(i, plt.cm.Set1(i/hl_max))
                    for i in hl_labels])
        else:
            if isinstance(hl_colors, list):
                hl_colors_out = dict(zip(hl_labels, hl_colors))
            elif isinstance(hl_colors, dict):
                for key in hl_colors.keys():
                    if key not in hl_labels:
                        raise BLE(ValueError(
                            'hl_colors dict must have an entry '
                            'for each hl_label.'))
                hl_colors_out = hl_colors
            else:
                raise BLE(TypeError('hl_colors must be a list or dict.'))
            if len(hl_colors) != len(hl_labels):
                raise BLE(ValueError('hl_colors must have an entry for each '
                    'entry in hl_labels.'))

        return hl_colors_out


    @staticmethod
    def gen_blank_sort(num_rows, num_cols):
        """Generate 'blank sort'.

        Allows state plotting of raw data without structure.
        """
        sorted_views = [0]
        sorted_clusters = {}
        sorted_clusters[0] = [0]
        sorted_rows = {}
        sorted_rows[0] = {}
        sorted_rows[0][0] = range(num_rows)
        sorted_cols = {}
        sorted_cols[0] = range(num_cols)

        return sorted_views, sorted_clusters, sorted_cols, sorted_rows


    @staticmethod
    def gen_cell_colors(T, sorted_views, sorted_cols, sorted_clusters,
            sorted_rows, column_partition, cmap, border_width,
            nan_color=(1., 0., 0., 1.)):
        """Generate heatmap using the data

        Allows clusters to have different base colors.
        """
        num_rows = len(T)
        num_cols = len(T[0])
        num_views = len(sorted_views)

        base_cmapper = np.zeros((num_rows, num_cols+num_views*border_width))

        x_pos = 0
        for view_count, view in enumerate(sorted_views):
            for col in sorted_cols[view]:
                y_pos = 0
                for cluster in sorted_clusters[view]:
                    for row in sorted_rows[view][cluster]:
                        xposidx = x_pos + view_count*border_width
                        base_cmapper[y_pos, xposidx] = float(cluster)
                        y_pos += 1
                x_pos += 1

        base_cmapper /= np.max(base_cmapper)
        cell_colors = np.zeros((base_cmapper.shape[0], base_cmapper.shape[1], 4))
        col_indices = []
        for view in sorted_views:
            idx = sorted_cols[view]
            col_indices += idx + [-1]*border_width
        for x_pos, col in enumerate(col_indices):
            if col < 0:
                cell_colors[:, col, :] = np.array([1, 1, 1, 1])
                continue

            view = column_partition[col]
            col_cpy = np.array([[x] if x is not None else [float('NaN')]
                                for x in T[:, col]])
            col_cpy = col_cpy[np.isfinite(col_cpy)]
            cmin = np.min(col_cpy)
            cmax = np.max(col_cpy)
            y_pos = 0
            for clstr in sorted_clusters[view]:
                for row in sorted_rows[view][clstr]:
                    value = T[row, col]
                    base_color = cmap(base_cmapper[y_pos, x_pos])
                    color = DrawStateUtils.cmap_color_brightness(value,
                        base_color, cmin, cmax, nan_color=nan_color)
                    cell_colors[y_pos, x_pos, :] = color
                    y_pos += 1

        return cell_colors



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

class IBayesDBForeignPredictorFactory(object):
    """PRELIMINARY BayesDB foreign predictor factory interface.

    A foreign predictor factory knows how to create instances of
    particular kind of foreign predictor, either by training them from
    data or by loading them from their serialized representation.

    Note: A common implementation pattern is to make the class of the
    foreign predictor be a singleton factory by defining these methods
    as `@classmethod`, instead of creating a separate class.  See, for
    example, :class:`bdbcontrib.predictors.keplers_law.KeplersLaw`.
    """

    def name(self):
        """Return the name of this kind of foreign predictor, as a string.

        All registered foreign predictors must have unique names.
        """
        raise NotImplementedError

    def create(self, bdb, table, targets, conditions):
        """Create and train a foreign predictor for the given circumstances.

        The `targets` and `conditions` ultimately come from the schema
        the client indicates.

        TODO `create` is currently assumed to be deterministic.  To
        support FPs whose training is stochastic, the :class:`.Composer`
        metamodel will need to be extended with options for
        maintaining ensembles of independently trained instances of
        such FPs.

        Parameters
        ----------
        bdb : :class:`bayeslite.BayesDB`
            The BayesDB containing the data to train on.

        table : string
            The name of the BayesDB table containing the data.

        targets : list<tuple<string, string>>
            The columns to be predicted, as pairs of column name and
            stattype.
            TODO Until :class:`.Composer` supports joint prediction,
            the `targets` list will have length 1.

        conditions : list<tuple<string, string>>
            The columns to be used as inputs, as pairs of column name and
            stattype.

        Returns
        -------
        predictor : :class:`~.IBayesDBForeignPredictor`
            A trained instance of the foreign predictor.

        See Also
        --------

        :func:`bdbcontrib.bql_utils.table_to_df`
            a helper function you can use if you want to train from a
            Pandas DataFrame rather than interacting directly with the
            database.
        """
        raise NotImplementedError

    def serialize(self, bdb, predictor):
        """Serialize the given predictor instance to a string.

        The instance will have been created by calling either
        :meth:`create` or :meth:`deserialize` on this factory.
        """
        raise NotImplementedError

    def deserialize(self, bdb, blob):
        """Reconstitute a serialized predictor instance.

        The `blob` will have been created by calling :meth:`serialize`
        on an instance of :class:`IBayesDBForeignPredictorFactory`
        registered with the same :meth:`name` as this one at some
        point in the past.  In typical use, this will be the same
        instance as the present one, but the usual advice about
        backward compatibility with stored data formats applies.
        """
        raise NotImplementedError

class IBayesDBForeignPredictor(object):
    """PRELIMINARY BayesDB trained foreign predictor instance interface.

    A foreign predictor (FP) is itself an independent object which is typically
    used outside the universe of BayesDB.

    Training is the province of
    :meth:`.IBayesDBForeignPredictorFactory.create`.

    A trained foreign predictor must support:

    - Simulation of the target column for a new observation

    - Evaluation of the logpdf of the target column of a new
      observation taking a certain value

    Simulate and logpdf both require a full realization of all `conditions`.

    TODO We do not yet support referring to row-specific training
    results or querying predictive distributions for observed rows.

    BayesDB initializes foreign predictors through the methods in
    :class:`~IBayesDBForeignPredictorFactory`, so imposes no
    particular restrictions on `__init__`.
    """

    def simulate(self, n_samples, conditions):
        """Simulate from the distribution {`target`}|{`conditions`}.

        The distribution being simulated from implicitly depends upon
        the data through the results of training.

        Parameters
        ----------
        n_samples : int
            Number of samples to simulate.

        conditions : dict
            A dictionary of {'condition':'value'} for all `conditions`
            required by the FP.  The FP may signal an error if any
            `conditions` are missing, and should ignore any additional
            ones.

        Returns
        -------
        list
            A list of the simulated values.
        """
        raise NotImplementedError

    def logpdf(self, value, conditions):
        """Evaluate the log-density of {`target`=`value`}|{`conditions`}.

        The distribution being evaluated implicitly depends upon the
        data through the results of training.

        Parameters
        ----------
        value : object
            The value of `target` to query.  The type will appropriate
            to the statistical type of the column.

        conditions : dict
            A dictionary of {'condition':'value'} for all `conditions`
            required by the FP.  The FP may signal an error if any
            `conditions` are missing, and should ignore any additional
            ones.

        Returns
        -------
        float
            The log probability density of the given target value
            given the conditions.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function

# matplotlib needs to set the backend before anything else gets to.
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from collections import namedtuple
from contextlib import contextmanager
from io import BytesIO
from textwrap import dedent
import bayeslite
import numpy as np
import os
import pandas as pd
import pytest
import re
import sys
import test_plot_utils
from test_plot_utils import dts_df

from bayeslite.exception import BayesLiteException
from bayeslite.read_csv import bayesdb_read_csv
from bayeslite.exception import BayesLiteException as BLE
from bayeslite.loggers import CaptureLogger

def test_quick_describe_variables_and_column_type(dts_df):
    dts, _df = dts_df
    resultdf = dts.query('SELECT * from %t LIMIT 1')
    assert ['index', 'floats_1', 'categorical_1', 'categorical_2',
            'few_ints_3', 'floats_3', 'many_ints_4', 'skewed_numeric_5',
            ] == list(resultdf.columns)
    resultdf = dts.quick_describe_variables()
    expected = {'floats_1': 'numerical',
                'categorical_1': 'categorical',
                'categorical_2': 'categorical',
                'few_ints_3': 'categorical',
                'floats_3': 'numerical',
                'many_ints_4': 'numerical',
                'skewed_numeric_5': r'(categorical|numerical)'}
    for (column, expected_type) in expected.iteritems():
        stattype = resultdf[resultdf['name'] == column]['stattype'].iloc[0]
        assert re.match(expected_type, stattype), column
        assert re.match(expected_type, dts.vartype(column))

def test_explore_vars(dts_df):
    dts, _df = dts_df
    dts.logger.calls = []
    with pytest.raises(BLE):
        dts.quick_explore_vars([]) # Empty columns
    with pytest.raises(BLE):
        dts.quick_explore_vars(['float_1']) # Just one column.

    dts.quick_explore_vars(['floats_1', 'categorical_1'], plotfile='foo')
    call_types = pd.DataFrame([call[0] for call in dts.logger.calls])
    call_counts = call_types.iloc[:,0].value_counts()
    assert call_counts['result'] > 0
    assert call_counts['plot'] > 0
    assert 'warn' not in call_counts

def test_similar_rows(dts_df):
    dts, _df = dts_df
    dts.logger.calls = []
    the_real_queryfn = dts.query
    count = [0]
    q_calls = []
    results = [pd.DataFrame([[0]]),  # Count of matching rows.
               pd.DataFrame([[2]]),  # Count of matching rows.
               pd.DataFrame([[1]]),  # Count of matching rows.
               None,                 # Call with temp table creation.
               'similar rows',       # Similarity result.
               ]
    def my_test_queryfn(fmt_str, *args):
        q_calls.append((fmt_str, args))
        count[0] += 1
        return results[count[0] - 1]
    try:
        dts.query = my_test_queryfn
        with pytest.raises(BLE):
            dts.quick_similar_rows(identify_row_by={})  # No rows.
        assert 'SELECT COUNT(*)' == q_calls[-1][0][:15]
        with pytest.raises(BLE):
            dts.quick_similar_rows(identify_row_by={})  # Too many rows.
        assert 'SELECT COUNT(*)' == q_calls[-1][0][:15]
        assert 'similar rows' == dts.quick_similar_rows(identify_row_by={
            'a':'b', 'c': 'd'})
        assert 'SELECT COUNT(*)' == q_calls[-3][0][:15]
        assert 'CREATE TEMP TABLE' == q_calls[-2][0][:17]
        assert re.search(r'''"a" = [?]\s+and\s+"c" = [?] ''',
                         q_calls[-2][0])
        assert (['b', 'd'],) == q_calls[-2][1]
        assert 'SELECT * FROM' == q_calls[-1][0][:13]
    finally:
        dts.query = the_real_queryfn


    call_types = pd.DataFrame([call[0] for call in dts.logger.calls])
    if len(call_types) > 0:
        call_counts = call_types.iloc[:,0].value_counts()
        assert 'warn' not in call_counts



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sqlite3

import numpy as np

import bayeslite.core as core
from bayeslite.exception import BayesLiteException as BLE
from bayeslite.math_util import logmeanexp
from bayeslite.sqlite3_util import sqlite3_quote_name as quote
from bayeslite.util import casefold

from crosscat.utils import sample_utils as su

import bayeslite.metamodel

composer_schema_1 = [
'''
INSERT INTO bayesdb_metamodel
    (name, version) VALUES ('composer', 1);
''','''
CREATE TABLE bayesdb_composer_cc_id(
    generator_id INTEGER NOT NULL PRIMARY KEY
        REFERENCES bayesdb_generator(id),
    crosscat_generator_id INTEGER NOT NULL UNIQUE
        REFERENCES bayesdb_generator(id),

    CHECK (generator_id != crosscat_generator_id)
);
''','''
CREATE TABLE bayesdb_composer_column_owner(
    generator_id INTEGER NOT NULL REFERENCES bayesdb_generator(id),
    colno INTEGER NOT NULL,
    local BOOLEAN NOT NULL,

    PRIMARY KEY(generator_id, colno),
    FOREIGN KEY(generator_id, colno)
        REFERENCES bayesdb_generator_column(generator_id, colno)
);
''','''
CREATE TABLE bayesdb_composer_column_toposort(
    generator_id INTEGER NOT NULL REFERENCES bayesdb_generator(id),
    colno INTEGER NOT NULL,
        -- Cannot express check constraint with subquery in sqlite.
        -- CHECK (EXISTS(SELECT generator_id, colno
        --      FROM bayesdb_composer_column_owner WHERE local = FALSE)),
    position INTEGER NOT NULL,
        CHECK (0 <= position)

    PRIMARY KEY(generator_id, colno),
    FOREIGN KEY(generator_id, colno)
        REFERENCES bayesdb_generator_column(generator_id, colno),
    UNIQUE (generator_id, position)
);
''','''
CREATE TRIGGER bayesdb_composer_column_toposort_check
    BEFORE INSERT ON bayesdb_composer_column_toposort
BEGIN
    SELECT CASE
        WHEN (NOT EXISTS(SELECT generator_id, colno, local
                FROM bayesdb_composer_column_owner
                WHERE generator_id=NEW.generator_id AND colno=NEW.colno AND
                    local = 0))
        THEN RAISE(ABORT, 'Columns in bayesdb_composer_column_toposort
            must be foreign.')
    END;
END;
''','''
CREATE TABLE bayesdb_composer_column_parents(
    generator_id INTEGER NOT NULL REFERENCES bayesdb_generator(id),
    fcolno INTEGER NOT NULL,
    pcolno INTEGER NOT NULL,

    PRIMARY KEY(generator_id, fcolno, pcolno),
    FOREIGN KEY(generator_id, fcolno)
        REFERENCES bayesdb_generator_column(generator_id, colno),
    FOREIGN KEY(generator_id, pcolno)
        REFERENCES bayesdb_generator_column(generator_id, colno),
    CHECK (fcolno != pcolno)
);
''','''
CREATE TABLE bayesdb_composer_column_foreign_predictor(
    generator_id INTEGER NOT NULL REFERENCES bayesdb_generator(id),
    colno INTEGER NOT NULL,
    predictor_name TEXT COLLATE NOCASE NOT NULL,
    predictor_binary BLOB,

    PRIMARY KEY(generator_id, colno),
    FOREIGN KEY(generator_id, colno)
        REFERENCES bayesdb_generator_column(generator_id, colno)
);
''','''
CREATE TRIGGER bayesdb_composer_column_foreign_predictor_check
    BEFORE INSERT ON bayesdb_composer_column_foreign_predictor
BEGIN
    SELECT CASE
        WHEN (NOT EXISTS(SELECT generator_id, colno, local
                FROM bayesdb_composer_column_owner
                WHERE generator_id=NEW.generator_id AND colno=NEW.colno AND
                    local = 0))
        THEN RAISE(ABORT, 'Columns in bayesdb_composer_foreign_predictor
            must be foreign.')
    END;
END;
''']

class Composer(bayeslite.metamodel.IBayesDBMetamodel):
    """A metamodel which composes foreign predictors with CrossCat.
    """

    def __init__(self, n_samples=None):
        # In-memory map of registered foreign predictor builders.
        self.predictor_builder = {}
        self.predictor_cache = {}
        # Default number of samples.
        if n_samples is None:
            self.n_samples = 100
        else:
            assert 0 < n_samples
            self.n_samples = n_samples

    def _predictor_cache(self, bdb):
        assert bdb.cache is not None
        if 'composer' in bdb.cache:
            return bdb.cache['composer']
        else:
            comp_cache = {}
            bdb.cache['composer'] = comp_cache
            return comp_cache

    def register_foreign_predictor(self, builder):
        """Register an object which builds a foreign predictor.

        Explicitly initializing a foreign predictor instance is not
        necessary. The `composer` will create, train, and serialize
        all foreign predictors declared in the `schema` when the BQL query
        INITIALIZE is called.

        Foreign predictors must be registered each time the database is
        launched.

        Parameters
        ----------
        builder : :class:`.IBayesDBForeignPredictorFactory`
            The pattern used by the extant predictors is to include these
            four methods in the class implementing
            :class:`~.IBayesDBForeignPredictor` as `@classmethods`. For
            example, registering
            :class:`bdbcontrib.predictors.random_forest.RandomForest` is
            achieved by registering the **class** instance::

                >> from bdbcontrib.predictors.random_forest import RandomForest
                >> composer.register_foreign_predictor(RandomForest)
        """
        # Validate the builder.
        # Not isinstance(builder, predictor.IBayesDBForeignPreidctorFactory)
        # because the pattern using classes and class methods does not make the
        # classes be instances of that.
        assert hasattr(builder, 'create')
        assert hasattr(builder, 'serialize')
        assert hasattr(builder, 'deserialize')
        assert hasattr(builder, 'name')
        # Check for duplicates.
        if casefold(builder.name()) in self.predictor_builder:
            raise BLE(ValueError(
                'A foreign predictor with name "{}" has already '
                'been registered. Currently registered: {}'.format(
                    builder.name(), self.predictor_builder)))
        self.predictor_builder[casefold(builder.name())] = builder

    def name(self):
        return 'composer'

    def register(self, bdb):
        cursor = bdb.sql_execute('''
            SELECT version FROM bayesdb_metamodel WHERE name = ?;
        ''', (self.name(),))
        version = None
        try:
            row = cursor.next()
        except StopIteration:
            version = 0
        else:
            version = row[0]
        assert version is not None
        if version == 0:
            with bdb.savepoint():
                for stmt in composer_schema_1:
                    bdb.sql_execute(stmt)
        return

    def create_generator(self, bdb, table, schema, instantiate):
        # Parse the schema.
        (columns, lcols, _fcols, fcol_to_pcols, fcol_to_fpred,
            dependencies) = self.parse(schema)
        # Instantiate **this** generator.
        genid, bdbcolumns = instantiate(columns.items())
        # Create internal crosscat generator. The name will be the same as
        # this generator name, with a _cc suffix.
        SUFFIX = '_cc'
        cc_name = bayeslite.core.bayesdb_generator_name(bdb, genid) + SUFFIX
        # Create strings for crosscat schema.
        cc_cols = ','.join('{} {}'.format(quote(c), quote(columns[c]))
                           for c in lcols)
        cc_dep = []
        for dep, colnames in dependencies:
            qcns = ','.join(map(quote, colnames))
            if dep:
                cc_dep.append('DEPENDENT({})'.format(qcns))
            else:
                cc_dep.append('INDEPENDENT({})'.format(qcns))
        bql = """
            CREATE GENERATOR {} FOR {} USING crosscat(
                {}, {}
            );
        """.format(quote(cc_name), quote(table), cc_cols, ','.join(cc_dep))
        bdb.execute(bql)
        # Convert strings to column numbers.
        fcolno_to_pcolnos = {}
        for f in fcol_to_pcols:
            fcolno = core.bayesdb_generator_column_number(bdb, genid, f)
            fcolno_to_pcolnos[fcolno] = [core.bayesdb_generator_column_number(
                bdb, genid, col) for col in fcol_to_pcols[f]]
        with bdb.savepoint():
            # Save internal cc generator id.
            bdb.sql_execute('''
                INSERT INTO bayesdb_composer_cc_id
                    (generator_id, crosscat_generator_id) VALUES (?,?)
            ''', (genid, core.bayesdb_get_generator(bdb, cc_name),))
            # Save lcols/fcolnos.
            for colno, _, _ in bdbcolumns:
                local = colno not in fcolno_to_pcolnos
                bdb.sql_execute('''
                    INSERT INTO bayesdb_composer_column_owner
                        (generator_id, colno, local) VALUES (?,?,?)
                ''', (genid, colno, int(local),))
            # Save parents of foreign columns.
            for fcolno in fcolno_to_pcolnos:
                for pcolno in fcolno_to_pcolnos[fcolno]:
                    bdb.sql_execute('''
                        INSERT INTO bayesdb_composer_column_parents
                            (generator_id, fcolno, pcolno) VALUES (?,?,?)
                    ''', (genid, fcolno, pcolno,))
            # Save topological order.
            topo = self.topological_sort(fcolno_to_pcolnos)
            for position, (colno, _) in enumerate(topo):
                bdb.sql_execute('''
                    INSERT INTO bayesdb_composer_column_toposort
                        (generator_id, colno, position) VALUES (?,?,?)
                    ''', (genid, colno, position,))
            # Save predictor names of foreign columns.
            for fcolno in fcolno_to_pcolnos:
                fp_name = fcol_to_fpred[casefold(
                    core.bayesdb_generator_column_name(bdb,genid, fcolno))]
                bdb.sql_execute('''
                    INSERT INTO bayesdb_composer_column_foreign_predictor
                        (generator_id, colno, predictor_name) VALUES (?,?,?)
                ''', (genid, fcolno, casefold(fp_name)))

    def drop_generator(self, bdb, genid):
        with bdb.savepoint():
            # Clear caches.
            keys = [k for k in self._predictor_cache(bdb) if k[0] == genid]
            for k in keys:
                del self._predictor_cache(bdb)[k]
            # Obtain before losing references.
            cc_name = core.bayesdb_generator_name(bdb, self.cc_id(bdb, genid))
            # Delete tables reverse order of insertion.
            bdb.sql_execute('''
                DELETE FROM bayesdb_composer_column_foreign_predictor
                    WHERE generator_id = ?
            ''', (genid,))
            bdb.sql_execute('''
                DELETE FROM bayesdb_composer_cc_id
                    WHERE generator_id = ?
            ''', (genid,))
            bdb.sql_execute('''
                DELETE FROM bayesdb_composer_column_toposort
                    WHERE generator_id = ?
            ''', (genid,))
            bdb.sql_execute('''
                DELETE FROM bayesdb_composer_column_parents
                    WHERE generator_id = ?
            ''', (genid,))
            bdb.sql_execute('''
                DELETE FROM bayesdb_composer_column_owner
                    WHERE generator_id = ?
            ''', (genid,))
            # Drop internal crosscat.
            bdb.execute('''
                DROP GENERATOR {}
            '''.format(quote(cc_name)))

    def initialize_models(self, bdb, genid, modelnos, model_config):
        # Initialize internal crosscat, maintaining equality of model numbers.
        # The semantics of INITIALIZE are that it guarantees the existence
        # of a sequence of models up to the requested number of them,
        # and BayesDB computes the numbers that need to be filled in.
        # The inverse of that computation is max(modelnos)+1.
        qg = quote(core.bayesdb_generator_name(bdb, self.cc_id(bdb, genid)))
        bql = 'INITIALIZE {} MODELS FOR {};'.format(max(modelnos)+1, qg)
        bdb.execute(bql)
        # Initialize the foriegn predictors.
        for fcol in self.fcols(bdb, genid):
            # Convert column numbers to names.
            targets = \
                [(core.bayesdb_generator_column_name(bdb, genid, fcol),
                  core.bayesdb_generator_column_stattype(bdb, genid, fcol))]
            conditions = \
                [(core.bayesdb_generator_column_name(bdb, genid, pcol),
                  core.bayesdb_generator_column_stattype(bdb, genid, pcol))
                 for pcol in self.pcols(bdb, genid, fcol)]
            # Initialize the foreign predictor.
            table_name = core.bayesdb_generator_table(bdb, genid)
            predictor_name = self.predictor_name(bdb, genid, fcol)
            builder = self.predictor_builder[predictor_name]
            predictor = builder.create(bdb, table_name, targets, conditions)
            # Store in the database.
            with bdb.savepoint():
                sql = '''
                    UPDATE bayesdb_composer_column_foreign_predictor SET
                        predictor_binary = :predictor_binary
                        WHERE generator_id = :genid AND colno = :colno
                '''
                predictor_binary = builder.serialize(bdb, predictor)
                bdb.sql_execute(sql, {
                    'genid': genid,
                    'predictor_binary': sqlite3.Binary(predictor_binary),
                    'colno': fcol
                })

    def drop_models(self, bdb, genid, modelnos=None):
        qg = quote(core.bayesdb_generator_name(bdb, self.cc_id(bdb, genid)))
        if modelnos is not None:
            models = ",".join(str(modelno) for modelno in modelnos)
            bql = 'DROP MODELS {} FROM {};'.format(models, qg)
        else:
            bql = 'DROP MODELS FROM {};'.format(qg)
        bdb.execute(bql)

    def analyze_models(self, bdb, genid, modelnos=None, iterations=1,
                max_seconds=None, ckpt_iterations=None, ckpt_seconds=None):
        # XXX Composer currently does not perform joint inference.
        # (Need full GPM interface, active research project).
        self.cc(bdb, genid).analyze_models(bdb, self.cc_id(bdb, genid),
            modelnos=modelnos, iterations=iterations, max_seconds=max_seconds,
            ckpt_iterations=ckpt_iterations, ckpt_seconds=ckpt_seconds)
        # Accounting.
        sql = '''
            UPDATE bayesdb_generator_model
                SET iterations = iterations + :iterations
                WHERE generator_id = :generator_id AND modelno = :modelno
        '''
        if modelnos is None:
            modelnos = core.bayesdb_generator_modelnos(bdb, genid)
        with bdb.savepoint():
            for modelno in modelnos:
                bdb.sql_execute(sql, {
                    'generator_id': genid,
                    'modelno': modelno,
                    'iterations': iterations,
                })

    def column_dependence_probability(self, bdb, genid, modelno, colno0,
            colno1):
        if modelno is None:
            modelnos = core.bayesdb_generator_modelnos(bdb, genid)
        else:
            modelnos = [modelno]
        with bdb.savepoint():
            p = sum(self._column_dependence_probability(
                     bdb, genid, m, colno0, colno1)
                    for m in modelnos) / float(len(modelnos))
        return p

    def _column_dependence_probability(self, bdb, genid, modelno, colno0,
            colno1):
        # XXX Computes the dependence probability for a single model.
        if modelno is None:
            raise BLE(ValueError('Invalid modelno argument for '
                'internal _column_dependence_probability. An integer modelno '
                'is required, not None.'))
        # Trivial case.
        if colno0 == colno1:
            return 1
        fcols = set(self.fcols(bdb, genid))
        c0_foreign = colno0 in fcols
        c1_foreign = colno1 in fcols
        # Neither col is foreign, delegate to CrossCat.
        # XXX Fails for future implementation of conditional dependence.
        if not (c0_foreign or c1_foreign):
            return self.cc(bdb, genid).column_dependence_probability(bdb,
                self.cc_id(bdb, genid), modelno,
                self.cc_colno(bdb, genid, colno0),
                self.cc_colno(bdb, genid, colno1))
        # (colno0, colno1) form a (target, conditions) pair.
        # WE explicitly modeled them as dependent by assumption.
        # TODO: Strong assumption? What if FP determines it is not
        # dependent on one of its conditions? (ie 0 coeff in regression)
        if colno0 in self.pcols(bdb, genid, colno1) or \
                colno1 in self.pcols(bdb, genid, colno0):
            return 1
        # (colno0, colno1) form a (local, foreign) pair.
        # IF [col0 FP target], [col1 CC], and [all conditions of col0 IND col1]
        #   then [col0 IND col1].
        # XXX Reverse is not true generally (counterxample), but we shall
        # assume an IFF condition. This assumption is not unlike the transitive
        # closure property of independence in crosscat.
        if (c0_foreign and not c1_foreign) or \
                (c1_foreign and not c0_foreign):
            fcol = colno0 if c0_foreign else colno1
            lcol = colno1 if c0_foreign else colno0
            return any(self._column_dependence_probability(bdb, genid, modelno,
                pcol, lcol) for pcol in self.pcols(bdb, genid, fcol))
        # XXX TODO: Determine independence semantics for this case.
        # Both columns are foreign. (Recursively) return 1 if any of their
        # conditions (possibly FPs) have dependencies.
        assert c0_foreign and c1_foreign
        return any(self._column_dependence_probability(bdb, genid, modelno,
                pcol0, pcol1) for pcol0 in self.pcols(bdb, genid, colno0)
                for pcol1 in self.pcols(bdb, genid, colno1))

    def column_mutual_information(self, bdb, genid, modelno, colno0, colno1,
            numsamples=None):
        if numsamples is None:
            numsamples = self.n_samples
        # XXX Aggregator only.
        row_id = core.bayesdb_generator_fresh_row_id(bdb, genid)
        X = [(row_id, colno0)]
        W = [(row_id, colno1)]
        Z = Y = []
        if modelno is None:
            modelnos = core.bayesdb_generator_modelnos(bdb, genid)
        else:
            modelnos = [modelno]
        with bdb.savepoint():
            mi = sum(self.conditional_mutual_information(
                      bdb, genid, modelno, X, W, Z, Y)
                     for modelno in modelnos) / float(len(modelnos))
        return mi

    def conditional_mutual_information(self, bdb, genid, modelno, X, W, Z, Y,
            numsamples=None):
        with bdb.savepoint():
            return self._conditional_mutual_information(
                bdb, genid, modelno, X, W, Z, Y, numsamples=numsamples)

    def _conditional_mutual_information(self, bdb, genid, modelno, X, W, Z, Y,
            numsamples=None):
        # WARNING: SUPER EXPERIMENTAL.
        # Computes the conditional mutual information I(X:W|Z,Y=y), defined
        # defined as the expectation E_z~Z{X:W|Z=z,Y=y}.
        # X, W, and Z must each be a list [(rowid, colno), ..].
        # Y is an evidence list [(rowid,colno,val), ..].
        if numsamples is None:
            numsamples = self.n_samples
        # All sets must be disjoint.
        all_cols = X + W + Z + [(r,c) for r,c,_ in Y]
        if len(all_cols) != len(set(all_cols)):
            raise BLE(ValueError('Duplicate cells received in '
                'conditional_mutual_information.\n'
                'X: {}\nW: {}\nZ: {}\nY: {}'.format(X, W, Z, Y)))
        # Simulate from joint.
        XWZ_samples = self.simulate(bdb, genid, modelno, X+W+Z,
            Y, numpredictions=numsamples)
        # Simple Monte Carlo
        mi = logpz = logpxwz = logpxz = logpwz = 0
        for s in XWZ_samples:
            Qx = [(r,c,v) for ((r,c),v) in zip(X, s[:len(X)])]
            Qw = [(r,c,v) for ((r,c),v) in zip(W, s[len(X):len(X)+len(W)])]
            Qz = [(r,c,v) for ((r,c),v) in zip(Z, s[len(X)+len(W):])]
            if Z:
                logpz = self._joint_logpdf(bdb, genid, modelno, Qz, Y)
            else:
                logpz = 0
            logpxwz = self._joint_logpdf(bdb, genid, modelno, Qx+Qw+Qz, Y)
            logpxz = self._joint_logpdf(bdb, genid, modelno, Qx+Qz, Y)
            logpwz = self._joint_logpdf(bdb, genid, modelno, Qw+Qz, Y)
            mi += logpz + logpxwz - logpxz - logpwz
        # TODO: linfoot?
        # TODO: If negative, report to user that reliable answer cannot be
        # returned with current `numsamples`.
        # Averaging is in direct space is correct.
        return mi/numsamples

    def logpdf_joint(self, bdb, generator_id, targets, constraints, modelno):
        if modelno is None:
            modelnos = core.bayesdb_generator_modelnos(bdb, generator_id)
        else:
            modelnos = [modelno]
        with bdb.savepoint():
            return logmeanexp([self._joint_logpdf(bdb, generator_id, modelno,
                targets, constraints) for modelno in modelnos])

    def _joint_logpdf(self, bdb, genid, modelno, Q, Y, n_samples=None):
        # XXX Computes the joint probability of query Q given evidence Y
        # for a single model. The function is a likelihood weighted
        # integrator.
        # XXX Determine.
        if n_samples is None:
            n_samples = self.n_samples
        # Validate inputs.
        if modelno is None:
            raise BLE(ValueError('Invalid modelno None, integer requried.'))
        if len(Q) == 0:
            raise BLE(ValueError('Invalid query Q: len(Q) == 0.'))
        # Ensure consistency of any duplicates in Q and Y.
        Q = self._queries_consistent_with_constraints(Q, Y)
        if Q is None:
            return float('-inf')
        for r, _, _ in Q+Y:
            assert r == Q[0][0], "Cannot assess more than one row, "\
                "%s and %s requested" % (Q[0][0], r)
        # (Q,Y) marginal joint density.
        _, QY_weights = self._weighted_sample(bdb, genid, modelno,
            Q[0][0], Q+Y, n_samples=n_samples)
        # Y marginal density.
        _, Y_weights = self._weighted_sample(bdb, genid, modelno,
            Q[0][0], Y, n_samples=n_samples)
        # XXX TODO Keep sampling until logpQY <= logpY
        logpQY = logmeanexp(QY_weights)
        logpY = logmeanexp(Y_weights)
        return logpQY - logpY

    def _queries_consistent_with_constraints(self, Q, Y):
        queries = dict()
        for (row, col, val) in Q:
            if (row, col) in queries:
                raise BLE(ValueError('Cannot specify duplicate query columns.'))
            queries[(row, col)] = val
        ignore = set()
        constraints = set()
        for (row, col, val) in Y:
            if (row, col) in constraints:
                raise BLE(ValueError(
                    'Cannot specify duplicate constraint row, column.'))
            if (row, col) in queries:
                if queries[(row, col)] == val:
                    ignore.add((row, col))
                else:
                    return None
            constraints.add((row, col))
        return [(r, c, v) for r,c,v in Q if (r,c) not in ignore]

    def predict_confidence(self, bdb, genid, modelno, colno, rowid,
            numsamples=None):
        with bdb.savepoint():
            return self._predict_confidence(bdb, genid, modelno, colno, rowid,
                numsamples=numsamples)

    def _predict_confidence(self, bdb, genid, modelno, colno, rowid,
            numsamples=None):
        # Predicts a value for the cell [rowid, colno] with a confidence metric.
        # XXX Prefer accuracy over speed for imputation.
        if numsamples is None:
            numsamples = self.n_samples
        colnos = core.bayesdb_generator_column_numbers(bdb, genid)
        colnames = core.bayesdb_generator_column_names(bdb, genid)
        row = core.bayesdb_generator_row_values(bdb, genid, rowid)
        # Account for multiple imputations if imputing parents.
        parent_conf = 1
        # Predicting lcol.
        if colno in self.lcols(bdb, genid):
            # Delegate to CC IFF
            # (lcol has no children OR all its children are None).
            children = [f for f in self.fcols(bdb, genid) if colno in
                    self.pcols(bdb, genid, f)]
            if len(children) == 0 or \
                    all(row[i] is None for i in xrange(len(row)) if i+1
                        in children):
                return self.cc(bdb, genid).predict_confidence(bdb,
                        self.cc_id(bdb, genid), modelno,
                        self.cc_colno(bdb, genid, colno), rowid)
            else:
                # Obtain likelihood weighted samples from posterior.
                Q = [(rowid, colno)]
                Y = [(rowid, c, v) for c,v in zip(colnos, row)
                     if c != colno and v is not None]
                samples = self.simulate(bdb, genid, modelno, Q, Y,
                    numpredictions=numsamples)
                samples = [s[0] for s in samples]
        # Predicting fcol.
        else:
            conditions = {c:v for c,v in zip(colnames, row) if
                core.bayesdb_generator_column_number(bdb, genid, c) in
                self.pcols(bdb, genid, colno)}
            for colname, val in conditions.iteritems():
                # Impute all missing parents.
                if val is None:
                    imp_col = core.bayesdb_generator_column_number(bdb, genid,
                        colname)
                    imp_val, imp_conf = self.predict_confidence(bdb, genid,
                        modelno, imp_col, rowid, numsamples=numsamples)
                    # XXX If imputing several parents, take the overall
                    # overall conf as min conf. If we define imp_conf as
                    # P[imp_val = correct] then we might choose to multiply
                    # the imp_confs, but we cannot assert that the imp_confs
                    # are independent so multiplying is extremely conservative.
                    parent_conf = min(parent_conf, imp_conf)
                    conditions[colname] = imp_val
            assert all(v is not None for c,v in conditions.iteritems())
            predictor = self.predictor(bdb, genid, colno)
            samples = predictor.simulate(numsamples, conditions)
        # Since foreign predictor does not know how to impute, imputation
        # shall occur here in the composer by simulate/logpdf calls.
        stattype = core.bayesdb_generator_column_stattype(bdb, genid, colno)
        if stattype == 'categorical':
            # imp_conf is most frequent.
            imp_val =  max(((val, samples.count(val)) for val in set(samples)),
                key=lambda v: v[1])[0]
            if colno in self.fcols(bdb, genid):
                imp_conf = np.exp(predictor.logpdf(imp_val, conditions))
            else:
                imp_conf = sum(np.array(samples)==imp_val) / len(samples)
        elif stattype == 'numerical':
            # XXX The definition of confidence is P[k=1] where
            # k=1 is the number of mixture componets (we need a distribution
            # over GPMM to answer this question). The confidence is instead
            # implemented as \max_i{p_i} where p_i are the weights of a
            # fitted DPGMM.
            imp_val = np.mean(samples)
            imp_conf = su.continuous_imputation_confidence(samples, None, None,
                n_steps=1000)
        else:
            raise BLE(ValueError(
                'Unknown stattype "{}" for a foreign predictor '
                'column encountered in predict_confidence.'.format(stattype)))
        return imp_val, imp_conf * parent_conf

    def simulate_joint(self, bdb, generator_id, targets, constraints, modelno,
            num_predictions=1):
        with bdb.savepoint():
            return self.simulate(bdb, generator_id, modelno, targets,
                constraints, numpredictions=num_predictions)

    def simulate(self, bdb, genid, modelno, targets, constraints,
            numpredictions=1):
        # Delegate to crosscat if colnos+constraints all lcols.
        colnos = [c for _,c in targets]
        all_cols = [c for _,c,_ in constraints] + colnos
        if all(f not in all_cols for f in self.fcols(bdb, genid)):
            Y_cc = [(r, self.cc_colno(bdb, genid, c), v)
                for r, c, v in constraints]
            Q_cc = [(r, self.cc_colno(bdb, genid, c)) for r,c in targets]
            return self.cc(bdb, genid).simulate_joint(bdb,
                self.cc_id(bdb, genid), Q_cc, Y_cc, modelno,
                num_predictions=numpredictions)
        # Solve inference problem by sampling-importance resampling.
        result = []
        for r,_ in targets:
            assert r == targets[0][0], "Cannot simulate more than one row, "\
                "%s and %s requested" % (targets[0][0], r)
        for _ in xrange(numpredictions):
            samples, weights = self._weighted_sample(bdb, genid, modelno,
                targets[0][0], constraints)
            p = np.exp(np.asarray(weights) - np.max(weights))
            p /= np.sum(p)
            draw = np.nonzero(bdb.np_prng.multinomial(1,p))[0][0]
            s = [samples[draw].get(col) for col in colnos]
            result.append(s)
        return result

    def row_similarity(self, bdb, genid, modelno, rowid, target_rowid,
            colnos):
        # XXX Delegate to CrossCat always.
        cc_colnos = self.cc_colnos(bdb, genid, colnos)
        return self.cc(bdb, genid).row_similarity(bdb, self.cc_id(bdb, genid),
            modelno, rowid, target_rowid, cc_colnos)

    def _weighted_sample(self, bdb, genid, modelno, row_id, Y, n_samples=None):
        # Returns a pairs of parallel lists ([sample ...], [weight ...])
        # Each `sample` is a dict {col:v} of values for all nodes in
        # the network for one row. Y specifies evidence nodes as (row,
        # col, value) triples: all returned samples have constrained
        # values at the evidence nodes.
        # `weight` is the likelihood of the evidence Y under s\Y.
        if n_samples is None:
            n_samples = self.n_samples
        # Create n_samples dicts, each entry is weighted sample from joint.
        samples = [{c:v for r,c,v in Y if r == row_id}
                   for _ in xrange(n_samples)]
        weights = []
        w0 = 0
        # Assess likelihood of evidence at root.
        Y_cc = [(r, c, v) for r,c,v in Y if c in self.lcols(bdb, genid)]
        if Y_cc:
            w0 += self.cc(bdb, genid).logpdf_joint(bdb, self.cc_id(bdb, genid),
                Y_cc, [], modelno)
        # Simulate unobserved ccs.
        Q_cc = [(row_id, c)
                for c in self.lcols(bdb, genid) if c not in samples[0]]
        V_cc = self.cc(bdb, genid).simulate_joint(bdb, self.cc_id(bdb, genid),
            Q_cc, Y_cc, modelno, num_predictions=n_samples)
        for k in xrange(n_samples):
            w = w0
            # Add simulated Q_cc.
            samples[k].update({c:v for (_, c), v in zip(Q_cc, V_cc[k])})
            for fcol in self.topo(bdb, genid):
                pcols = self.pcols(bdb, genid, fcol)
                predictor = self.predictor(bdb, genid, fcol)
                # All parents of FP known (evidence or simulated)?
                assert pcols.issubset(set(samples[k]))
                conditions = {core.bayesdb_generator_column_name(
                    bdb, genid, c):v for c,v in samples[k].iteritems()
                        if c in pcols}
                if fcol in samples[k]:
                    # f is evidence: compute likelihood weight.
                    w += predictor.logpdf(samples[k][fcol], conditions)
                else:
                    # f is latent: simulate from conditional distribution.
                    samples[k][fcol] = predictor.simulate(1, conditions)[0]
            weights.append(w)
        return samples, weights

    def cc_colno(self, bdb, genid, colno):
        return self.cc_colnos(bdb, genid, [colno])[0]

    def cc_colnos(self, bdb, genid, colnos):
        lcolnames = [bayeslite.core.bayesdb_generator_column_name(bdb,
            genid, colno) for colno in colnos]
        return [bayeslite.core.bayesdb_generator_column_number(bdb,
            self.cc_id(bdb, genid), lcolname) for lcolname in lcolnames]

    def cc_id(self, bdb, genid):
        cursor = bdb.sql_execute('''
            SELECT crosscat_generator_id FROM bayesdb_composer_cc_id
                WHERE generator_id = ?
        ''', (genid,))
        return cursor.fetchall()[0][0]

    def cc(self, bdb, genid):
        return core.bayesdb_generator_metamodel(bdb, self.cc_id(bdb, genid))

    def lcols(self, bdb, genid):
        cursor = bdb.sql_execute('''
            SELECT colno FROM bayesdb_composer_column_owner
                WHERE generator_id = ? AND local = 1
                ORDER BY colno ASC
        ''', (genid,))
        return set([row[0] for row in cursor])

    def fcols(self, bdb, genid):
        cursor = bdb.sql_execute('''
            SELECT colno FROM bayesdb_composer_column_owner
                WHERE generator_id = ? AND local = 0
                ORDER BY colno ASC
        ''', (genid,))
        return set([row[0] for row in cursor])

    def pcols(self, bdb, genid, fcolno):
        cursor = bdb.sql_execute('''
            SELECT pcolno FROM bayesdb_composer_column_parents
                WHERE generator_id = ? AND fcolno = ?
                ORDER BY pcolno ASC
        ''', (genid, fcolno))
        return set([row[0] for row in cursor])

    def topo(self, bdb, genid):
        cursor = bdb.sql_execute('''
            SELECT colno FROM bayesdb_composer_column_toposort
                WHERE generator_id = ?
                ORDER BY position ASC
            ''', (genid,))
        return [row[0] for row in cursor]

    def predictor_name(self, bdb, genid, fcol):
        cursor = bdb.sql_execute('''
            SELECT predictor_name FROM bayesdb_composer_column_foreign_predictor
                WHERE generator_id = ? AND colno = ?
        ''', (genid, fcol))
        return cursor.fetchall()[0][0]

    def predictor(self, bdb, genid, fcol):
        if (genid, fcol) not in self._predictor_cache(bdb):
            cursor = bdb.sql_execute('''
                SELECT predictor_name, predictor_binary
                    FROM bayesdb_composer_column_foreign_predictor
                    WHERE generator_id = ? AND colno = ?
            ''', (genid, fcol))
            name, binary = cursor.fetchall()[0]
            builder = self.predictor_builder.get(name, None)
            if builder is None:
                raise BLE(LookupError('Foreign predictor for column "{}" '
                    'not registered: "{}".'.format(name,
                        core.bayesdb_generator_column_name(bdb, genid, fcol))))
            self._predictor_cache(bdb)[(genid, fcol)] = \
                builder.deserialize(bdb, binary)
        return self._predictor_cache(bdb)[(genid, fcol)]

    def parse(self, schema):
        """Parse the given `schema` for a `composer` metamodel.

        An example of a schema is::

            CREATE GENERATOR foo FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner
                    CATEGORICAL, Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL, Apogee_km
                    NUMERICAL, Eccentricity NUMERICAL, Launch_Mass_kg NUMERICAL,
                    Dry_Mass_kg NUMERICAL, Power_watts NUMERICAL, Date_of_Launch
                    NUMERICAL, Anticipated_Lifetime NUMERICAL, Contractor
                    CATEGORICAL, Country_of_Contractor CATEGORICAL, Launch_Site
                    CATEGORICAL, Launch_Vehicle CATEGORICAL,
                    Source_Used_for_Orbital_Data CATEGORICAL,
                    longitude_radians_of_geo NUMERICAL, Inclination_radians
                    NUMERICAL
                ),
                random_forest (
                    Type_of_Orbit CATEGORICAL
                        GIVEN Apogee_km, Perigee_km,
                            Eccentricity, Period_minutes, Launch_Mass_kg,
                            Power_watts, Anticipated_Lifetime, Class_of_orbit
                ),
                keplers_law (
                    Period_minutes NUMERICAL
                        GIVEN Perigee_km, Apogee_km
                ),
                multiple_regression (
                    Anticipated_Lifetime NUMERICAL
                        GIVEN Dry_Mass_kg, Launch_Mass_kg, Purpose
                ),
                dependent(Launch_Mass_kg, Dry_Mass_kg, Power_watts),
                dependent(Perigee_km, Apogee_km),
                independent(Operator_Owner, Inclination_radians)
            );

        The schema must adhere to the following rules:

        - Default metamodel is identified `default` or
          `crosscat`. Every `colname` must have its `stattype`
          declared. IGNORE and GUESS(*) are forbidden.

        - Foriegn predictors are identified by the `name()` method of
          the object used when `Composer.register_foreign_predictor`
          was invoked.

        For example::

            >> from bdbcontrib.foreign.random_forest import RandomForest
            >> composer.register_foreign_predictor(random_forest.RandomForest)
            >> RandomForest.name()
            random_forest

        The grammar inside foreign predictor directives is::

            <target> <stattype> GIVEN <condition> [...[condition]]

        All columns specified in `dependent` and `independent`
        directives must be modeled by the `default` metamodel.

        Parameters
        ----------
        schema : list<list>
            The `schema` as parsed by bayesdb.

        Returns
        -------
        columns : dict(str:str)
            A dict(colname:stattype) mapping every `colname` declared in
            `schema` to its `stattype`.

        lcols : list<str>
            A list of columns modeled by `default` model.

        fcols : list<str>
            A list of columns modeled by foreign predictor.

        fcol_to_pcols : dict(str:list<str>)
            A dict(fcol:conditions) mapping `fcol` to a list of its parent
            columns.

        fcol_to_fpred : dict(str:str)
            A dict(fcol:fpred) mapping `fcol` to the name of its foreign
            predictor. The values in the dictionary are keys in
            `self.predictor_builder`.

        dependencies : list(<tuple(<bool>,<list<str>)>)
            A list of dependency constraints. Each entry in the list
            is a tuple.  For example, (True, ['foo', 'bar', 'baz'])
            means the three variables are mutually and pairwise
            *dependent*.

        """
        # Allowed keywords.
        DIRECTIVES = ['crosscat', 'default', 'dependent', 'independent'] + \
            self.predictor_builder.keys()
        STATTYPES = ['numerical', 'categorical']
        # Data structures to return.
        columns = {}
        lcols = []
        fcols = []
        fcol_to_pcols = dict()
        fcol_to_fpred = dict()
        dependencies = []
        # Parse!
        for block in schema:
            if len(block) == 0:
                continue
            directive = casefold(block[0])
            commands = block[1]
            if directive not in DIRECTIVES:
                raise BLE(ValueError('Unknown directive "{}".\n'
                    'Available directives: {}.'.format(directive, DIRECTIVES)))
            if not isinstance(commands, list):
                raise BLE(ValueError('Unknown commands in "{}" directive: {}.'\
                        .format(directive, commands)))
            if directive == 'default' or directive == 'crosscat':
                while commands:
                    c = casefold(commands.pop(0))
                    if c == ',':
                        continue
                    s = casefold(commands.pop(0))
                    if s not in STATTYPES:
                        raise BLE(ValueError(
                            'Invalid stattype "{}".'.format(s)))
                    columns[c] = s
                    lcols.append(c)
            elif directive == 'independent':
                ind = []
                while commands:
                    c = casefold(commands.pop(0))
                    if c == ',':
                        continue
                    ind.append(c)
                dependencies.append((False, ind))
            elif directive == 'dependent':
                dep = []
                while commands:
                    c = casefold(commands.pop(0))
                    if c == ',':
                        continue
                    dep.append(c)
                dependencies.append((True, dep))
            elif directive in self.predictor_builder:
                c = casefold(commands.pop(0))
                s = casefold(commands.pop(0))
                if s not in STATTYPES:
                    raise BLE(ValueError('Invalid stattype "{}".'.format(s)))
                columns[c] = s
                given = casefold(commands.pop(0))
                if given != 'given':
                    raise BLE(ValueError(
                        'Execpted GIVEN keyword, received: {}.'.format(given)))
                conditions = []
                while commands:
                    r = casefold(commands.pop(0))
                    if r == ',':
                        continue
                    conditions.append(r)
                fcols.append(c)
                fcol_to_pcols[c] = conditions
                fcol_to_fpred[c] = directive
        # Unique lcols.
        if len(lcols) != len(set(lcols)):
            raise BLE(ValueError(
                'Duplicate default columns enountered: {}.'.format(lcols)))
        # Unique fcols.
        if len(fcols) != len(set(fcols)):
            raise BLE(ValueError(
                'Duplicate foreign columns enountered: {}.'.format(fcols)))
        # All stattypes declared.
        for _, c in fcol_to_pcols.iteritems():
            for r in c:
                if r not in columns:
                    raise BLE(ValueError(
                        'No stattype declaration for "{}".'.format(r)))
        # No col both lcol and fcol.
        for l in lcols:
            if l in fcol_to_pcols:
                raise BLE(ValueError(
                    'Column "{}" can only be modeled once.'.format(l)))
        # No non-default dependencies.
        for dep in dependencies:
            for col in dep[1]:
                if col not in lcols:
                    raise BLE(ValueError(
                        'Column "{}" with dependency constraint '
                        'must have default model.'.format(col)))
        # Return the hodgepodge.
        return (columns, lcols, fcols, fcol_to_pcols, fcol_to_fpred,
            dependencies)

    @staticmethod
    def topological_sort(graph):
        """Topologically sort a directed graph represented as an adjacency list.

        Assumes that edges are incoming, ie (10:[8,7]) means 8->10 and 7->10.

        Parameters
        ----------
        graph : list or dict
            Adjacency list or dict representing the graph, for example:
                graph_l = [(10, [8, 7]), (5, [8, 7, 9, 10, 11, 13, 15])]
                graph_d = {10: [8, 7], 5: [8, 7, 9, 10, 11, 13, 15]}

        Returns
        -------
        graph_sorted : list
            An adjacency list, where the order of the nodes is listed
            in topological order.
        """
        # Algorithm: until the unsorted graph is empty, find a node
        # that has no remaining predecessors, put it at the end of the
        # current sort, and remove it from the graph.
        # XXX This is O(v*(v+e)) in the worst case, whereas Tarjan's
        # alogrithm is O(v+e).  Tarjan's algorithm also reports the
        # location of cycles with greater precision if there happens
        # to be more than one.  Someone^TM should replace this with
        # that.
        graph_sorted = []
        graph = dict(graph)
        while graph:
            acyclic = False
            for node, edges in graph.items():
                for edge in edges:
                    if edge in graph:
                        break
                else:
                    acyclic = True
                    del graph[node]
                    graph_sorted.append((node, edges))
            if not acyclic:
                raise BLE(ValueError(
                    'A cyclic dependency occurred in topological_sort.'))
        return graph_sorted



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import shlex

import matplotlib
import matplotlib.pyplot as plt

from bayeslite.shell.hook import bayesdb_shell_cmd

import bdbcontrib
from bdbcontrib.shell_utils import ArgparseError, ArgumentParser


matplotlib.rcParams.update({'figure.autolayout': True,
    'font.weight': 'bold',
    'figure.facecolor': 'white'})


@bayesdb_shell_cmd('mihist')
def mi_hist(self, argin):
    """plot histogram of mutual information over generator's models
    <generator> <col1> <col2> [<options>]

    Example:
    bayeslite> .mihist sat_generator dry_mass launch_mass -n=1000
    """
    parser = ArgumentParser(prog='.mihist')
    parser.add_argument('generator', type=str, help='generator name')
    parser.add_argument('col1', type=str, help='first column')
    parser.add_argument('col2', type=str, help='second column')
    parser.add_argument('-n', '--num-samples', type=int, default=1000)
    parser.add_argument('-b', '--bins', type=int, default=15,
                        help='number of bins')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')

    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    fig = bdbcontrib.mi_hist(self._bdb, args.generator, args.col1,
        args.col2, num_samples=args.num_samples, bins=args.bins)

    if args.filename is None:
        plt.show()
    else:
        fig.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('heatmap')
def heatmap(self, argin):
    """plot clustered heatmap of pairwise matrix
    <query> [<options>]

    Options
    -------
    -f, --filename: str
        Save as filename
    --vmin: float
        Minimun value of the colormap.
    --vmax: float
        Maximum value of the colormap.

    Example:
    bayeslite> .heatmap ESTIMATE DEPENDENCE PROBABILITY FROM PAIRWISE COLUMNS OF mytable
    """

    parser = ArgumentParser(prog='.heatmap')
    parser.add_argument('bql', type=str, nargs='+', help='PAIRWISE BQL query')
    parser.add_argument('--vmin', type=float, default=None)
    parser.add_argument('--vmax', type=float, default=None)
    #XXX last-sort has been removed in favor of specifying row/col ordering.
    parser.add_argument('--last-sort', action='store_true')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')

    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bql = " ".join(args.bql)
    clustermap = bdbcontrib.heatmap(self._bdb, bql, vmin=args.vmin,
        vmax=args.vmax)

    if args.filename is None:
        plt.show()
    else:
        clustermap.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('show')
def pairplot(self, argin):
    """plot array of plots for all pairs of columns
    <query> [<options>]

    - continuous-continuous as scatter (optional KDE contour)
    - continuous-categorical as violinplot
    - categorical-categorical as heatmap

    Options:
        -f, --filename: the output filename. If not specified, tries
                        to draw.
        -g, --generator: the generator name. Providing the generator
                         name make help to plot more intelligently.
        -s, --shortnames: Use column short names to label facets?
        -m, --show-missing: Plot missing values in scatter plots as lines.
        -t, --no-tril: Plot the entire square matrix, not just lower triangle.
        --show-contour: Turn on contours.
        --colorby: The name of a column to use as a marker variable for color.

    Example:
    bayeslite> .show SELECT foo, baz, quux + glorb FROM mytable
        --filename myfile.png
    bayeslite> .show ESTIMATE foo, baz FROM mytable_cc -g mytable_cc
    """
    parser = ArgumentParser(prog='.show')
    parser.add_argument('bql', type=str, nargs='+', help='BQL query')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')
    parser.add_argument('-g', '--generator', type=str, default=None,
                        help='Query generator name')
    parser.add_argument('-s', '--shortnames', action='store_true',
                        help='Use column short names?')
    parser.add_argument('--no-contour', action='store_true',
                        help='Turn off countours (KDE).'),
    parser.add_argument('--show-contour', action='store_true',
                        help='Turn on contours (KDE).')
    parser.add_argument('-m', '--show-missing', action='store_true',
                        help='Plot missing values in scatterplot.')
    parser.add_argument('--show-full', action='store_true')
    parser.add_argument('--colorby', type=str, default=None,
                        help='Name of column to use as a dummy variable.')
    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bql  = " ".join(args.bql)
    figure = bdbcontrib.pairplot(self._bdb, bql,
        generator_name=args.generator, show_contour=args.show_contour,
        colorby=args.colorby, show_missing=args.show_missing,
        show_full=args.show_full)

    if args.filename is None:
        plt.show()
    else:
        figure.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('drawcc')
def draw_crosscat(self, argin):
    """plot crosscat state
    <generator> <modelno> [<options>]

    Options
        -f, --filename: the output filename. If not specified, tries to draw.

    Example:
    bayeslite> .ccstate mytable_cc 12 -f state_12.png
    """
    parser = ArgumentParser(prog='.drawcc')
    parser.add_argument('generator', type=str, help='Generator')
    parser.add_argument('modelno', type=int, help='Model number to plot.')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')
    parser.add_argument('-r', '--row-label-col', type=str, default=None,
                        help='The name of the column to use for row labels.')
    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    figure = bdbcontrib.draw_crosscat(self._bdb, args.generator,
        args.modelno, row_label_col=args.row_label_col)

    if args.filename is None:
        plt.show()
    else:
        figure.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('histogram')
def histogram(self, argin):
    """plot histogram of one- or two-column table
    <query> [<options>]

    If two-column, subdivide the first column according to labels in
    the second column.

    Example (plots overlapping histograms of height for males and females):
    bayeslite> .histogram SELECT height, sex FROM humans; --normed --bin 31
    """
    parser = ArgumentParser(prog='.histogram')
    parser.add_argument('bql', type=str, nargs='+', help='BQL query')
    parser.add_argument('-b', '--bins', type=int, default=15,
                        help='number of bins')
    parser.add_argument('--normed', action='store_true',
                        help='Normalize histograms?')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')
    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bql = " ".join(args.bql)
    figure = bdbcontrib.histogram(self._bdb, bql, args.bins,
        args.normed)

    if args.filename is None:
        plt.show()
    else:
        figure.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('bar')
def barplot(self, argin):
    """plot bar-plot of query giving categories and heights
    <query> [<options>]

    First column specifies names; second column specifies heights.
    """
    parser = ArgumentParser(prog='.bar')
    parser.add_argument('bql', type=str, nargs='+', help='BQL query')
    parser.add_argument('-f', '--filename', type=str, default=None,
                        help='output filename')
    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bql = " ".join(args.bql)
    figure = bdbcontrib.barplot(self._bdb, bql)

    if args.filename is None:
        plt.show()
    else:
        figure.savefig(args.filename)
    plt.close('all')


@bayesdb_shell_cmd('chainplot')
def plot_crosscat_chain_diagnostics(self, argin):
    """plot crosscat diagnostic for all models of generator
    <diagnostic> <generator> [output_filename]

    Valid (crosscat) diagnostics are:
        - logscore: log score of the model
        - num_views: the number of views in the model
        - column_crp_alpha: CRP alpha over columns

    Example:
    bayeslite> .chainplot logscore dha_cc scoreplot.png
    """
    parser = ArgumentParser(prog='.chainplot')
    parser.add_argument('diagnostic', type=str, help='Diagnostic name')
    parser.add_argument('generator', type=str, help='Generator name.')
    parser.add_argument('-f', '--filename', type=str, default=None,
        help='output filename')
    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    figure = bdbcontrib.plot_crosscat_chain_diagnostics(self._bdb,
        args.diagnostic, args.generator)

    if args.filename is None:
        plt.show()
    else:
        figure.savefig(args.filename)
    plt.close('all')



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import shlex

from bayeslite.shell.hook import bayesdb_shell_cmd

import bdbcontrib
from bdbcontrib.shell_utils import ArgparseError, ArgumentParser


@bayesdb_shell_cmd('est_ll')
def estimate_log_likelihood(self, argin):
    """estimate log likelihood for a dataset
    <table> <generator> [--targets-cols <...>] [--given-cols <...>] [--n-samples <N>]

    Examples:
    bayeslite> .est_ll people people_gen --targets height --givens age 1 nationality 17 --n-samples 1000
    """
    parser = ArgumentParser(prog='.est_ll')
    parser.add_argument('table', type=str,
        help='Name of the table.')
    parser.add_argument('generator', type=str,
        help='Name of the generator.')
    parser.add_argument('--targets', nargs='*',
        help='Sequence of target columns to evaluate the log likelhood. '
        'By default, all columns in <table> will be used.')
    parser.add_argument('--givens', nargs='*',
        help='Sequence of columns and observed values to condition on. '
        'The required format is [<col> <val>...].')
    parser.add_argument('--n-samples', type=int,
        help='Number of rows in the dataset to use in the computation. '
        'Defaults to all rows.')

    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    ll = bdbcontrib.estimate_log_likelihood(self._bdb, args.table,
        args.generator, targets=args.targets, givens=args.givens,
        n_samples=args.n_samples)

    print ll


@bayesdb_shell_cmd('est_kl')
def estimate_kl_divergence(self, argin):
    """estimate KL divergence of generator from reference
    <table> <generator-a> <generator-b> [--targets-cols <...>] [--given-cols <...>] [--n-samples <N>]

    Examples:
    bayeslite> .est_kl crosscat baxcat --targets height --givens age 1 nationality 17 --n-samples 1000
    """
    parser = ArgumentParser(prog='.est_kl')
    parser.add_argument('generator_a', type=str,
        help='Name of the reference generator.')
    parser.add_argument('generator_b', type=str,
        help='Name of the approximating generator.')
    parser.add_argument('--targets', nargs='*',
        help='Sequence of target columns to evaluate the log likelhood. '
        'By default, all columns in <table> will be used.')
    parser.add_argument('--givens', nargs='*',
        help='Sequence of columns and observed values to condition on. '
        'The required format is [<col> <val>...].')
    parser.add_argument('--n-samples', type=int,
        help='Number of rows in the dataset to use in the computation. '
        'Defaults to all rows.')

    try:
        args = parser.parse_args(shlex.split(argin))
    except ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    kl = bdbcontrib.estimate_kl_divergence(self._bdb, args.generator_a,
        args.generator_b, targets=args.targets, givens=args.givens,
        n_samples=args.n_samples)

    print kl



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import argparse
import cPickle as pickle # json doesn't like tuple dict keys
import logging
import math
import os

from bdbcontrib.experiments.probe import log, probe_fileset

######################################################################
## Probes                                                          ##
######################################################################

def country_purpose_probes(bdb):
    bdb.execute('''
        CREATE TEMP TABLE satellite_purpose AS
        SIMULATE country_of_operator, purpose FROM satellites_cc
        GIVEN Class_of_orbit = 'GEO', Dry_mass_kg = 500
        LIMIT 1000;
    ''')
    usa_ct_sql = '''
        SELECT COUNT(*) FROM satellite_purpose
        WHERE country_of_operator = 'USA'
          AND purpose = 'Communications'
    '''
    (usa_ct,) = bdb.execute(usa_ct_sql).next()
    hist_sql = '''
        SELECT country_of_operator, purpose, COUNT(*) AS frequency
        FROM satellite_purpose
        GROUP BY country_of_operator, purpose
        ORDER BY frequency DESC
        LIMIT 2
    '''
    results = bdb.execute(hist_sql).fetchall()
    usa_top = results[0][0] == 'USA' and results[0][1] == 'Communications'
    by_much = results[0][2] > 2*results[1][2]
    return [ ("USA-Communications count", "num", usa_ct),
             ("USA-Communications is top", "bool", usa_top),
             ("USA-Communications is top by 2x", "bool", by_much),
         ]

def unlikely_periods_probes(bdb):
    bdb.execute('''
        CREATE TEMP TABLE unlikely_periods AS
        ESTIMATE name, class_of_orbit, period_minutes,
        PREDICTIVE PROBABILITY OF period_minutes
            AS "Relative Probability of Period"
        FROM satellites_cc;
    ''')
    def prob_of(name):
        query = '''
            SELECT "Relative Probability of Period"
            FROM unlikely_periods
            WHERE name LIKE '%s'
        '''
        (ans,) = bdb.sql_execute(query % name).next()
        return ans
    top_ten = bdb.execute('''
        SELECT name FROM unlikely_periods
        WHERE class_of_orbit = 'GEO' AND period_minutes IS NOT NULL
        ORDER BY "Relative Probability of Period" ASC LIMIT 10;
    ''').fetchall()
    def in_top_ten(name):
        return any((n.startswith(name) for (n,) in top_ten))
    def in_top_five(name):
        return any((n.startswith(name) for (n,) in top_ten[0:5]))
    return [ ("DSP 20 period prob",    'num', prob_of('DSP 20 (USA 149)%')),
             ("SDS III-6 period prob", 'num', prob_of('SDS III-6%')),
             ("SDS III-7 period prob", 'num', prob_of('SDS III-7%')),
             ("Orion 6 period prob",   'num', prob_of('Advanced Orion 6%')),
             ("DSP 20 in top 10 unlikely periods",
              'bool', in_top_ten('DSP 20 (USA 149)')),
             ("SDS III-6 in top 10 unlikely periods",
              'bool', in_top_ten('SDS III-6')),
             ("SDS III-7 in top 10 unlikely periods",
              'bool', in_top_ten('SDS III-7')),
             ("Orion 6 in top 10 unlikely periods",
              'bool', in_top_ten('Advanced Orion 6')),
             ("DSP 20 in top 5 unlikely periods",
              'bool', in_top_five('DSP 20 (USA 149)')),
             ("SDS III-6 in top 5 unlikely periods",
              'bool', in_top_five('SDS III-6')),
             ("SDS III-7 in top 5 unlikely periods",
              'bool', in_top_five('SDS III-7')),
             ("Orion 6 in top 5 unlikely periods",
              'bool', in_top_five('Advanced Orion 6')),
         ]

def orbit_type_imputation_probes(bdb):
    bdb.execute('''
        CREATE TEMP TABLE inferred_orbit AS
        INFER EXPLICIT
        anticipated_lifetime, perigee_km, period_minutes, class_of_orbit,
        PREDICT type_of_orbit
            AS inferred_orbit_type
            CONFIDENCE inferred_orbit_type_conf
        FROM satellites_cc
        WHERE type_of_orbit IS NULL;
    ''')
    query = '''
        SELECT i.class_of_orbit, i.inferred_orbit_type,
            COUNT(*) AS count, AVG(i.inferred_orbit_type_conf) AS mean,
            AVG((i.inferred_orbit_type_conf - sub.answer) * (i.inferred_orbit_type_conf - sub.answer)) as variance
        FROM inferred_orbit AS i,
            (SELECT i.class_of_orbit, i.inferred_orbit_type,
                AVG(i.inferred_orbit_type_conf) AS answer
            FROM inferred_orbit as i
            GROUP BY i.class_of_orbit, i.inferred_orbit_type) AS sub
        WHERE i.class_of_orbit == sub.class_of_orbit
            AND i.inferred_orbit_type == sub.inferred_orbit_type
        GROUP BY i.class_of_orbit, i.inferred_orbit_type
        ORDER BY i.class_of_orbit, i.inferred_orbit_type
    '''
    answers = dict(((cl, tp), (ct, avg, var))
                   for (cl, tp, ct, avg, var)
                   in bdb.execute(query).fetchall())
    def probe_gen():
        for cl in ["Elliptical", "LEO"]:
            for tp in ["Deep Highly Eccentric",
                       "Intermediate",
                       "Molniya",
                       "Sun-Synchronous",
                       "Cislunar",
                       "N/A"]:
                (ct, avg, var) = answers.get((cl, tp), (0,0,0))
                yield ("%s %s ct imputed instances" % (cl, tp),
                       'num', float(ct))
                yield ("%s %s mean inference confidence" % (cl, tp),
                       'num', avg)
                yield ("%s %s inference confidence stddev" % (cl, tp),
                       'num', math.sqrt(var))
    return list(probe_gen())

######################################################################
## Driver                                                           ##
######################################################################

def doit(files, outfile, model_schedule, n_replications):
    out_dir = os.path.dirname(outfile)
    if out_dir and not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    results = probe_fileset(
        files, "satellites_cc",
        [country_purpose_probes,
         unlikely_periods_probes,
         orbit_type_imputation_probes],
        model_schedule = model_schedule,
        n_replications = n_replications)

    with open(outfile, "w") as f:
        pickle.dump(results, f)
    log("Saved probe results to %s" % outfile)

parser = argparse.ArgumentParser(
    description='Probe a collection of Satellites .bdb files')
parser.add_argument('-o', '--outfile',
                    help="Save results to the given file")
parser.add_argument('-m', '--n_models', nargs="+", type=int,
                    help="Count of models to probe")
parser.add_argument('-n', '--n_replications', type=int,
                    help="Number of replications to probe")
parser.add_argument('files', nargs="+", help=".bdb files to probe")

def main():
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)
    doit(args.files, args.outfile, args.n_models, args.n_replications)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math
import pickle
import numpy as np

import bdbcontrib
from bdbcontrib.predictors import predictor
from bayeslite.exception import BayesLiteException as BLE

class KeplersLaw(predictor.IBayesDBForeignPredictor):
    """A foreign predictor which models Kepler's Third Law.

    There must be exactly one `targets` column (period in minutes) and
    exactly two `conditions` columns (apogee in km, perigee in
    km, in that order). All stattypes are expected to be numerical.
    """

    @classmethod
    def create(cls, bdb, table, targets, conditions):
        cols = [c for c,_ in targets+conditions]
        df = bdbcontrib.bql_utils.table_to_df(bdb, table, cols)
        kl = cls()
        kl.train(df, targets, conditions)
        kl.prng = bdb.np_prng
        return kl

    @classmethod
    def serialize(cls, _bdb, pred):
        state = {
            'targets': pred.targets,
            'conditions': pred.conditions,
            'noise': pred.noise
        }
        return pickle.dumps(state)

    @classmethod
    def deserialize(cls, bdb, binary):
        state = pickle.loads(binary)
        kl = cls(targets=state['targets'], conditions=state['conditions'],
            noise=state['noise'])
        kl.prng = bdb.np_prng
        return kl

    @classmethod
    def name(cls):
        return 'keplers_law'

    def __init__(self, targets=None, conditions=None, noise=None):
        self.targets = targets
        self.conditions = conditions
        self.noise = noise

    def train(self, df, targets, conditions):
        # Obtain the targets column.
        if len(targets) != 1:
            raise BLE(ValueError('OrbitalMechanics can only target one '
                'column. Received {}'.format(targets)))
        if targets[0][1].lower() != 'numerical':
            raise BLE(ValueError('OrbitalMechanics can only target a NUMERICAL '
                'column. Received {}'.format(targets)))
        self.targets = [targets[0][0]]

        # Obtain the conditions column.
        if len(conditions) != 2:
            raise BLE(ValueError('OrbitalMechanics can only condition on '
                'two columns. Received {}'.format(conditions)))
        if any(c[1].lower() != 'numerical' for c in conditions):
            raise BLE(ValueError('OrbitalMechanics can only condition on '
                'NUMERICAL columns. Received {}'.format(conditions)))
        self.conditions = [c[0] for c in conditions]

        # The dataset.
        self.dataset = df[self.conditions + self.targets].dropna()
        X = self.dataset[self.conditions].as_matrix()

        # Learn the noise model.
        actual_period = self.dataset[self.targets].as_matrix().ravel()
        theoretical_period = satellite_period_minutes(X[:,0], X[:,1])
        errors = np.abs(actual_period-theoretical_period)
        error_95 = np.percentile(errors, 95)
        errors = np.mean(np.select([errors < error_95], [errors]))
        self.noise = np.sqrt(np.mean(errors**2))

    def _conditions(self, conditions):
        apogee_km_i = self.conditions[0]
        perigee_km_i = self.conditions[1]
        return conditions[apogee_km_i], conditions[perigee_km_i]

    def simulate(self, n_samples, conditions):
        if not set(self.conditions).issubset(set(conditions.keys())):
            raise BLE(ValueError(
                'Must specify values for all the conditionals.\n'
                'Received: {}\n'
                'Expected: {}'.format(conditions, self.conditions)))
        apogee_km, perigee_km = self._conditions(conditions)
        period_minutes = satellite_period_minutes(apogee_km, perigee_km)
        return list(period_minutes + self.prng.normal(scale=self.noise,
            size=n_samples))

    def logpdf(self, value, conditions):
        if not set(self.conditions).issubset(set(conditions.keys())):
            raise BLE(ValueError(
                'Must specify values for all the conditionals.\n'
                'Received: {}\n'
                'Expected: {}'.format(conditions, self.conditions)))
        apogee_km, perigee_km = self._conditions(conditions)
        period_minutes = satellite_period_minutes(apogee_km, perigee_km)
        return logpdfGaussian(value, period_minutes, self.noise)

HALF_LOG2PI = 0.5 * math.log(2 * math.pi)
def logpdfGaussian(x, mu, sigma):
    deviation = x - mu
    return - math.log(sigma) - HALF_LOG2PI \
        - (0.5 * deviation * deviation / (sigma * sigma))

def satellite_period_minutes(apogee_km, perigee_km):
    """Period of satellite with specified apogee and perigee.

    Apogee and perigee are in kilometers; period is in minutes."""
    GM = 398600.4418
    EARTH_RADIUS = 6378
    a = 0.5*(abs(apogee_km) + abs(perigee_km)) + EARTH_RADIUS
    return 2 * np.pi * np.sqrt(a**3/GM) / 60.
